--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- . Calculate the cross- and auto correlations of c_nof_inputs inputs.
-- Description:
-- . 

ENTITY correlator IS
  GENERIC (
    g_nof_input_streams  : NATURAL;
    g_nof_input_folds    : NATURAL := 0;  -- 2**g_nof_input_folds = number of inputs carried on one input stream
    g_nof_pre_mult_folds : NATURAL := 0;  -- Number of pre-multiplier stage folds.
    g_data_w             : NATURAL := 16; -- Complex data width
    g_conjugate          : BOOLEAN := TRUE;
    g_nof_channels       : NATURAL := 64; -- Number of (serial) channels per (parallel) input
    g_integration_period : NATURAL := 0   -- Expressed as the number of samples per channel to accumulate 
   ); 
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_input_streams-1 DOWNTO 0);

    src_out_arr    : OUT t_dp_sosi_arr(1-1 DOWNTO 0) -- Single output for now
  );
END correlator;

ARCHITECTURE str OF correlator IS

  CONSTANT c_nof_inputs                      : NATURAL := g_nof_input_streams*pow2(g_nof_input_folds);
  CONSTANT c_nof_visibilities                : NATURAL := (c_nof_inputs*(c_nof_inputs+1))/2;

  -----------------------------------------------------------------------------
  -- The array of multiplier input streams can be folded, e.g.:
  -- . c_nof_visibilities=10, g_nof_pre_mult_folds=0 -> c_nof_mults=10
  -- . c_nof_visibilities=10, g_nof_pre_mult_folds=1 -> c_nof_mults= 5
  -- . c_nof_visibilities=10, g_nof_pre_mult_folds=2 -> c_nof_mults= 3
  -- . c_nof_visibilities=10, g_nof_pre_mult_folds=3 -> c_nof_mults= 2
  -- . c_nof_visibilities=10, g_nof_pre_mult_folds=4 -> c_nof_mults= 1
  -- . c_nof_visibilities=10, g_nof_pre_mult_folds<0 -> c_nof_mults= 1
  -----------------------------------------------------------------------------
  CONSTANT c_nof_mults                       : NATURAL := sel_a_b(g_nof_pre_mult_folds>=0, ceil_div(c_nof_visibilities, ceil_pow2(g_nof_pre_mult_folds)), 1);

  -- We can multiplex the accumulator outputs onto one stream as long as the integration period is
  -- equal to or larger than the number of accumulator outputs.
  CONSTANT c_integration_period              : NATURAL := largest(g_integration_period, c_nof_mults);

  -- Nof accumulator sums to keep per stream depends on folding
  CONSTANT c_nof_accumulators                : NATURAL := g_nof_channels*pow2(g_nof_pre_mult_folds);

  CONSTANT c_mult_output_w                   : NATURAL := 2*g_data_w;
  CONSTANT c_acc_data_w                      : NATURAL := c_mult_output_w + ceil_log2(c_integration_period);

  SIGNAL corr_unfolder_src_out_arr           : t_dp_sosi_arr(c_nof_inputs-1 DOWNTO 0);
  SIGNAL corr_permutator_snk_in_arr          : t_dp_sosi_arr(c_nof_inputs-1 DOWNTO 0);
  SIGNAL corr_permutator_src_out_2arr_2      : t_dp_sosi_2arr_2(c_nof_visibilities-1 DOWNTO 0); -- Array of pairs
  SIGNAL corr_folder_snk_in_2arr_2           : t_dp_sosi_2arr_2(c_nof_visibilities-1 DOWNTO 0); -- Array of pairs, not folded yet
  SIGNAL corr_folder_src_out_2arr_2          : t_dp_sosi_2arr_2(c_nof_mults-1 DOWNTO 0);        -- Array of pairs, folded 
  SIGNAL corr_multiplier_snk_in_2arr_2       : t_dp_sosi_2arr_2(c_nof_mults-1 DOWNTO 0);
  SIGNAL corr_multiplier_src_out_arr         : t_dp_sosi_arr(c_nof_mults-1 DOWNTO 0);
  SIGNAL corr_accumulator_snk_in_arr         : t_dp_sosi_arr(c_nof_mults-1 DOWNTO 0);
  SIGNAL corr_accumulator_src_out_arr        : t_dp_sosi_arr(c_nof_mults-1 DOWNTO 0);
  SIGNAL corr_visibility_buffer_snk_in_arr   : t_dp_sosi_arr(c_nof_mults-1 DOWNTO 0);
  SIGNAL corr_visibility_buffer_src_out_arr  : t_dp_sosi_arr(c_nof_mults-1 DOWNTO 0);
  SIGNAL corr_folder_src_out_arr             : t_dp_sosi_arr(1-1 DOWNTO 0);

BEGIN

  ------------------------------------------------------------------------------
  -- Unfold input streams
  ------------------------------------------------------------------------------
  u_corr_unfolder : ENTITY work.corr_unfolder
  GENERIC MAP (
    g_nof_inputs  => g_nof_input_streams,
    g_nof_unfolds => g_nof_input_folds 
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,

    snk_in_arr  => snk_in_arr,

    src_out_arr => corr_unfolder_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Unfolder -> permutator pipeline to re-align the unfolder output
  -----------------------------------------------------------------------------
  gen_dp_pipeline_corr_unfolder_src_out_arr : FOR i IN 0 TO c_nof_inputs-1 GENERATE
    u_dp_pipeline : ENTITY dp_lib.dp_pipeline
    GENERIC MAP (
      g_pipeline => 0 + (pow2(g_nof_input_folds) - i REM pow2(g_nof_input_folds)-1)
    )
    PORT MAP (
      rst         => rst,
      clk         => clk,
  
      snk_in      => corr_unfolder_src_out_arr(i),
      src_out     => corr_permutator_snk_in_arr(i)
    );
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Create all unique pair permutations of the input streams
  -----------------------------------------------------------------------------
  u_corr_permutator : ENTITY work.corr_permutator
  GENERIC MAP (
    g_nof_inputs => c_nof_inputs
  )
  PORT MAP (
    clk            => clk,
    rst            => rst,

    snk_in_arr     => corr_permutator_snk_in_arr, --corr_unfolder_src_out_arr, --snk_in_arr,
    src_out_2arr_2 => corr_permutator_src_out_2arr_2
  );     

  -----------------------------------------------------------------------------
  -- Permutator -> folder pipeline
  -- . In addition to the normal pipelining, this pipeline stage also provides
  --   the correct timing of valid data for the folder stage.
  --   . The delay required for each stream i is : (i rem 2^nof_folds)
  --     . Example for 4 streams, fold 2 times:
  --                                           
  --       0) [0][1][2][3] -> pipeline -> [0]      [1]      [2]      [3]      
  --       1) [0][1][2][3] -> pipeline ->   [0]      [1]      [2]      [3]    
  --       2) [0][1][2][3] -> pipeline ->     [0]      [1]      [2]      [3]  
  --       3) [0][1][2][3] -> pipeline ->       [0]      [1]      [2]      [3]
  -----------------------------------------------------------------------------
  gen_dp_pipeline_corr_permutator_src_out_2arr2 : FOR i IN 0 TO c_nof_visibilities-1 GENERATE
    gen_dp_pipeline : FOR j IN 0 TO 2-1 GENERATE
      u_dp_pipeline : ENTITY dp_lib.dp_pipeline
      GENERIC MAP (
        g_pipeline => 0 + (i REM pow2(g_nof_pre_mult_folds))
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
    
        snk_in      => corr_permutator_src_out_2arr_2(i)(j),
        src_out     => corr_folder_snk_in_2arr_2(i)(j)
      );
    END GENERATE;
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Fold the streams before feeding them to the multiplier stage 
  -----------------------------------------------------------------------------
  u_corr_folder_2arr_2 : ENTITY work.corr_folder_2arr_2
  GENERIC MAP (
    g_nof_inputs => c_nof_visibilities,
    g_nof_folds  => g_nof_pre_mult_folds 
  )
  PORT MAP (
    rst            => rst,
    clk            => clk,

    snk_in_2arr_2  => corr_folder_snk_in_2arr_2,

    src_out_2arr_2 => corr_folder_src_out_2arr_2
  );

  -- If we're folding an uneven number of input pairs, we'll get a 50% valid
  -- rate on the highest stream index. We don't want that (because 
  -- common_shiftram in corr_accumulator does not support holes in the data),
  -- so override the valid with corr_folder_src_out_2arr_2(0).valid in that case.
  p_override_valid: PROCESS(corr_folder_src_out_2arr_2)
  BEGIN
    corr_multiplier_snk_in_2arr_2 <= corr_folder_src_out_2arr_2;
    IF g_nof_pre_mult_folds/=0 AND (c_nof_visibilities REM 2 = 1) THEN
      corr_multiplier_snk_in_2arr_2(c_nof_mults-1)(0).valid <= corr_folder_src_out_2arr_2(0)(0).valid;
      corr_multiplier_snk_in_2arr_2(c_nof_mults-1)(1).valid <= corr_folder_src_out_2arr_2(0)(0).valid;
    END IF;
  END PROCESS;

  -----------------------------------------------------------------------------
  -- Complex multiplier stage
  -----------------------------------------------------------------------------
  u_corr_multiplier : ENTITY work.corr_multiplier
  GENERIC MAP (
    g_nof_inputs => c_nof_mults,
    g_data_w     => g_data_w,
    g_conjugate  => g_conjugate
  )
  PORT MAP (
    clk            => clk,
    rst            => rst,

    snk_in_2arr_2  => corr_multiplier_snk_in_2arr_2,
    src_out_arr    => corr_multiplier_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Multiplier -> Accumulator pipeline
  -----------------------------------------------------------------------------
  gen_dp_pipeline_corr_multiplier_src_out_arr : FOR i IN 0 TO c_nof_mults-1 GENERATE
    u_dp_pipeline : ENTITY dp_lib.dp_pipeline
    GENERIC MAP (
      g_pipeline => 0
    )
    PORT MAP (
      rst         => rst,
      clk         => clk,
  
      snk_in      => corr_multiplier_src_out_arr(i),
      src_out     => corr_accumulator_snk_in_arr(i)
    );
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Accumulator stage
  -----------------------------------------------------------------------------
  u_corr_accumulator : ENTITY work.corr_accumulator
  GENERIC MAP (
    g_nof_inputs         => c_nof_mults,
    g_nof_accumulators   => c_nof_accumulators,
    g_integration_period => c_integration_period,
    g_data_w             => 2*g_data_w -- Multiplier output data width
  )
  PORT MAP (
    clk            => clk,
    rst            => rst,

    snk_in_arr     => corr_accumulator_snk_in_arr,
    src_out_arr    => corr_accumulator_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Accumulator -> FIFO pipeline
  -----------------------------------------------------------------------------
  gen_dp_pipeline_corr_accumulator_src_out_arr : FOR i IN 0 TO c_nof_mults-1 GENERATE
    u_dp_pipeline : ENTITY dp_lib.dp_pipeline
    GENERIC MAP (
      g_pipeline => 0
    )
    PORT MAP (
      rst         => rst,
      clk         => clk,
  
      snk_in      => corr_accumulator_src_out_arr(i),
      src_out     => corr_visibility_buffer_snk_in_arr(i)
    );
  END GENERATE;

  ------------------------------------------------------------------------------
  -- Buffer the visibilities before the folding stage.
  -- . All accumulators output blocks of visibilities at the same time. We need
  --   to buffer the visibilities to source them to the folder at the right
  --   pace.
  ------------------------------------------------------------------------------
  u_corr_visibility_buffer : ENTITY work.corr_visibility_buffer
  GENERIC MAP (
    g_nof_inputs         => c_nof_mults,
    g_data_w             => c_acc_data_w,
    g_buffer_depth       => c_nof_accumulators,
    g_nof_pre_mult_folds => g_nof_pre_mult_folds,
    g_integration_period => c_integration_period
  )
  PORT MAP (
    clk            => clk,
    rst            => rst,

    snk_in_arr     => corr_visibility_buffer_snk_in_arr,
    src_out_arr    => corr_visibility_buffer_src_out_arr
  );

  ------------------------------------------------------------------------------
  -- Fold onto one stream
  ------------------------------------------------------------------------------
  u_corr_folder : ENTITY work.corr_folder
  GENERIC MAP (
    g_nof_inputs => c_nof_mults,
    g_nof_folds  => -1 
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,

    snk_in_arr  => corr_visibility_buffer_src_out_arr,

    src_out_arr => corr_folder_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Add SOP, EOP, BSN, Channel and sync to tag the correlator output
  -----------------------------------------------------------------------------
  u_corr_output_framer : ENTITY work.corr_output_framer
  GENERIC MAP (
    g_nof_inputs          => 1,
    g_nof_words_per_frame => c_nof_visibilities, --c_nof_mults*pow2(g_nof_pre_mult_folds),
    g_nof_channels        => g_nof_channels,
    g_nof_frames_per_sync => g_nof_channels
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,

    snk_in_arr  => corr_folder_src_out_arr,

    src_out_arr => src_out_arr
  );

END str;
