--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- This testbench tests the different type of DDR controllers: 
--
--                        - uphy_4g_800_master
--                        - uphy_4g_800_slave
--
-- The DUT can be selected, using the c_tech_ddr constants. 
--
-- Testbench is selftesting: 
--
-- > as 8
-- > run -all 
--

LIBRARY IEEE, technology_lib, tech_ddr_lib, common_lib, dp_lib, diagnostics_lib;                   
USE IEEE.STD_LOGIC_1164.ALL;    
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;

ENTITY tb_io_ddr IS
  GENERIC (
    g_technology            : NATURAL := c_tech_select_default;
    g_ctlr_ref_clk_period   : TIME := 5 ns;     -- 200 MHz
    g_dvr_clk_period        : TIME := 5 ns;    -- 50 ns
    g_dp_clk_period         : TIME := 5000 ps;  -- 200 MHz
    g_dp_data_w             : NATURAL := 32;    -- 32 for mixed width and 256 for equal width FIFO
    g_nof_repeat            : NATURAL := 2;
    g_wr_flush_mode         : STRING := "SYN"  -- "VAL", "SOP", "SYN"
  );
END ENTITY tb_io_ddr;

ARCHITECTURE str of tb_io_ddr IS

  CONSTANT c_cross_domain_dvr_ctlr : BOOLEAN := g_ctlr_ref_clk_period/=g_dvr_clk_period;

  CONSTANT c_tech_ddr         : t_c_tech_ddr := c_tech_ddr3_4g_800m_master;
  
--  CONSTANT c_ddr_addr_hi      : t_tech_ddr_addr := ((OTHERS=>'0'),
--                                                    (OTHERS=>'0'),
--                                                    TO_UVEC(                                         3, c_tech_ddr_max.a_row_w),
--                                                    TO_UVEC(2**c_tech_ddr_4g.a_col_w-c_tech_ddr_4g.rsl, c_tech_ddr_max.a_col_w));
  CONSTANT c_ddr_addr_hi      : t_tech_ddr_addr := ((OTHERS=>'0'),
                                                    (OTHERS=>'0'),
                                                    TO_UVEC(                                         3, c_tech_ddr_max.a_row_w),
                                                    TO_UVEC(2**c_tech_ddr_4g.a_col_w-1, c_tech_ddr_max.a_col_w));
  -- DDR access stimuli
  CONSTANT c_address_lo_arr   : t_nat_natural_arr := (  0, 517,    0,     1,   2,   1,   2,   3,   5,  1900, 1900);
  CONSTANT c_nof_address_arr  : t_nat_natural_arr := (517, 507, 1024,     1,   6,   1,   1,   2,   3,   153,  153);
  CONSTANT c_wr_not_rd_arr    : STD_LOGIC_VECTOR  := ('1', '1',  '0',   '1', '1', '0', '0', '0', '0',    '1', '0');
                                                    
  CONSTANT c_ctlr_data_w      : NATURAL := func_tech_ddr_ctlr_data_w(c_tech_ddr);
  CONSTANT c_dp_factor        : NATURAL := c_ctlr_data_w/g_dp_data_w;
  
  CONSTANT c_wr_fifo_depth    : NATURAL := 128;
  CONSTANT c_rd_fifo_depth    : NATURAL := 256;
  
  -- Frame size for sop/eop
  CONSTANT c_wr_frame_size    : NATURAL := 32;
  -- Sync period
  CONSTANT c_wr_sync_period   : NATURAL := 512;
 
  SIGNAL tb_end               : STD_LOGIC := '0';
  SIGNAL ctlr_ref_clk         : STD_LOGIC := '0';
  SIGNAL ctlr_ref_rst         : STD_LOGIC;
  SIGNAL ctlr_clk             : STD_LOGIC;
  SIGNAL ctlr_rst             : STD_LOGIC;
  SIGNAL dvr_clk              : STD_LOGIC := '0';
  SIGNAL dvr_rst              : STD_LOGIC;
  SIGNAL dp_clk               : STD_LOGIC := '0';
  SIGNAL dp_rst               : STD_LOGIC;

  SIGNAL dvr_start_addr       : t_tech_ddr_addr; 
  SIGNAL dvr_end_addr         : t_tech_ddr_addr;

  SIGNAL dvr_en               : STD_LOGIC;
  SIGNAL dvr_done             : STD_LOGIC;
  SIGNAL dvr_wr_not_rd        : STD_LOGIC;
  SIGNAL dvr_wr_flush_en      : STD_LOGIC;
  
  SIGNAL diag_wr_src_in       : t_dp_siso;
  SIGNAL diag_wr_src_out      : t_dp_sosi;
  
  SIGNAL wr_fifo_usedw        : STD_LOGIC_VECTOR(ceil_log2(c_wr_fifo_depth * c_dp_factor)-1 DOWNTO 0);    
  SIGNAL wr_src_out           : t_dp_sosi;
  SIGNAL wr_val_cnt           : NATURAL := 0;

  SIGNAL diag_rd_snk_out      : t_dp_siso;
  SIGNAL diag_rd_snk_in       : t_dp_sosi;
  SIGNAL rd_fifo_usedw        : STD_LOGIC_VECTOR(ceil_log2(c_rd_fifo_depth * c_dp_factor)-1 DOWNTO 0);    

  SIGNAL src_diag_en          : STD_LOGIC;
  SIGNAL src_val_cnt          : STD_LOGIC_VECTOR(31 DOWNTO 0);

  SIGNAL snk_diag_en          : STD_LOGIC;
  SIGNAL snk_diag_res         : STD_LOGIC;
  SIGNAL snk_diag_res_val     : STD_LOGIC;
  SIGNAL snk_val_cnt          : STD_LOGIC_VECTOR(31 DOWNTO 0);
  SIGNAL expected_cnt         : NATURAL;
  
  SIGNAL phy_in               : t_tech_ddr_phy_in;
  SIGNAL phy_io               : t_tech_ddr_phy_io;
  SIGNAL phy_ou               : t_tech_ddr_phy_ou;
  
BEGIN
 
  ctlr_ref_clk <= NOT ctlr_ref_clk OR tb_end AFTER g_ctlr_ref_clk_period/2; 
  ctlr_ref_rst <= '1', '0' AFTER 100 ns;
  
  dvr_clk  <= NOT dvr_clk OR tb_end AFTER g_dvr_clk_period/2; 
  dvr_rst  <= '1', '0' AFTER 100 ns;
  
  dp_clk   <= NOT dp_clk OR tb_end AFTER g_dp_clk_period/2; 
  dp_rst   <= '1', '0' AFTER 100 ns;

  p_stimuli : PROCESS
    VARIABLE v_addr_lo : t_tech_ddr_addr;
    VARIABLE v_addr_hi : t_tech_ddr_addr;
  BEGIN
    tb_end          <= '0';
    dvr_en          <= '0';
    dvr_wr_flush_en <= '0';
    dvr_wr_not_rd   <= '0';
    src_diag_en     <= '0';
    snk_diag_en     <= '0';
    expected_cnt    <= 0;

    proc_common_wait_until_high(dvr_clk, dvr_done);
    
    -- Start diagnostics source for write and sink for verify read
    src_diag_en <= '1';
    snk_diag_en <= '1';
    
    -- After reset the write FIFO is flushed until the first write access is started, even when dvr_wr_flush_en='0'
    proc_common_wait_some_cycles(ctlr_clk, 100);
    
    FOR R IN 0 TO g_nof_repeat-1 LOOP
      FOR I IN c_address_lo_arr'RANGE LOOP
        v_addr_lo := func_tech_ddr_dq_address(TO_DDR_CTLR_ADDRESS(4*c_address_lo_arr(I))                         , c_tech_ddr);
        v_addr_hi := func_tech_ddr_dq_address(TO_DDR_CTLR_ADDRESS(4*c_address_lo_arr(I)+4*c_nof_address_arr(I)-1), c_tech_ddr);
        
        dvr_start_addr <= v_addr_lo;
        dvr_end_addr   <= v_addr_hi;
        
        -- START ACCESS
        dvr_wr_not_rd <= c_wr_not_rd_arr(I);
        dvr_en        <= '1';
        proc_common_wait_some_cycles(dvr_clk, 1);
        dvr_en        <= '0'; 
        
        -- ACCESS DONE
        proc_common_wait_until_lo_hi(dvr_clk, dvr_done);
        
        IF c_wr_not_rd_arr(I)='0' THEN
          expected_cnt <= expected_cnt + c_nof_address_arr(I)*c_dp_factor;
        
          ASSERT snk_diag_res_val = '1' REPORT "[ERROR] DIAG_RES INVALID!"  SEVERITY FAILURE;
          ASSERT snk_diag_res = '0'     REPORT "[ERROR] NON-ZERO DIAG_RES!" SEVERITY FAILURE;
        END IF;
      END LOOP;
      
      -- Stop diagnostics source and sink
      src_diag_en <= '0';
      snk_diag_en <= '0';
      
      -- Flush the wr fifo
      dvr_wr_flush_en <= '1';
      proc_common_wait_some_cycles(dvr_clk, 1);
      dvr_wr_flush_en <= '0';
      
      -- Wait until the wr fifo has been flushed and the rd fifo has been read empty
      proc_common_wait_some_cycles(ctlr_clk, 500*c_dp_factor);
      ASSERT UNSIGNED(wr_fifo_usedw) = 0          REPORT "[ERROR] Write FIFO is flushed but not empty!" SEVERITY FAILURE;
      ASSERT UNSIGNED(rd_fifo_usedw) = 0          REPORT "[ERROR] Read FIFO is not empty!" SEVERITY FAILURE;
      ASSERT UNSIGNED(snk_val_cnt) = expected_cnt REPORT "[ERROR] Unexpected number of read data!" SEVERITY FAILURE;
      
      -- Restart diagnostics source and sink
      src_diag_en <= '1';
      snk_diag_en <= '1';
    END LOOP;
    
    REPORT "[OK] Test passed." SEVERITY NOTE;
    tb_end <= '1';

    WAIT;
  END PROCESS;

  u_diagnostics: ENTITY diagnostics_lib.diagnostics 
  GENERIC MAP (
    g_dat_w             => g_dp_data_w,
    g_nof_streams       => 1
     ) 
  PORT MAP (
    rst                 => dp_rst,
    clk                 => dp_clk,

    snk_out_arr(0)      => diag_rd_snk_out,
    snk_in_arr(0)       => diag_rd_snk_in,
    snk_diag_en(0)      => snk_diag_en,
    snk_diag_md(0)      => '1',
    snk_diag_res(0)     => snk_diag_res,
    snk_diag_res_val(0) => snk_diag_res_val,
    snk_val_cnt(0)      => snk_val_cnt,

    src_out_arr(0)      => diag_wr_src_out,
    src_in_arr(0)       => diag_wr_src_in,
    src_diag_en(0)      => src_diag_en,
    src_diag_md(0)      => '1',
    src_val_cnt(0)      => src_val_cnt
  );

  wr_val_cnt <= wr_val_cnt + 1 WHEN rising_edge(dp_clk) AND diag_wr_src_out.valid='1';
  
  p_sop_eop : PROCESS (diag_wr_src_out, wr_val_cnt)
  BEGIN
    -- Default, fits g_wr_flush_mode="VAL"
    wr_src_out <= diag_wr_src_out;
    
    IF g_wr_flush_mode="SOP" THEN
      wr_src_out.sop <= '0';
      wr_src_out.eop <= '0';
      IF wr_val_cnt MOD c_wr_frame_size = 0 THEN
        wr_src_out.sop <= diag_wr_src_out.valid;
      ELSIF wr_val_cnt MOD c_wr_frame_size = c_wr_frame_size-1 THEN
        wr_src_out.eop <= diag_wr_src_out.valid;
      END IF;
    END IF;
    
    IF g_wr_flush_mode="SYN" THEN
      wr_src_out.sync <= '0';
      IF wr_val_cnt MOD c_wr_sync_period = 0 THEN
        wr_src_out.sync <= diag_wr_src_out.valid;
      END IF;
    END IF;
  END PROCESS;
    
  u_io_ddr: ENTITY work.io_ddr
  GENERIC MAP(
    g_technology             => g_technology,
    g_tech_ddr               => c_tech_ddr,
    g_cross_domain_dvr_ctlr  => c_cross_domain_dvr_ctlr,
    g_wr_data_w              => g_dp_data_w,
    g_wr_fifo_depth          => c_wr_fifo_depth,  -- >=16 AND >g_tech_ddr.maxburstsize, defined at DDR side of the FIFO.
    g_rd_fifo_depth          => c_rd_fifo_depth,  -- >=16 AND >g_tech_ddr.maxburstsize, defined at DDR side of the FIFO. 
    g_rd_data_w              => g_dp_data_w,
    g_wr_flush_mode          => g_wr_flush_mode,
    g_wr_flush_use_channel   => FALSE,
    g_wr_flush_start_channel => 0,
    g_wr_flush_nof_channels  => 1
  )                      
  PORT MAP (
    -- DDR reference clock
    ctlr_ref_clk       => ctlr_ref_clk,
    ctlr_ref_rst       => ctlr_ref_rst,
                                     
    -- DDR controller clock domain
    ctlr_clk_out       => ctlr_clk,
    ctlr_rst_out       => ctlr_rst,
    
    ctlr_clk_in        => ctlr_clk,  -- connect ctlr_clk_out to ctlr_clk_in at top level to avoid potential delta-cycle differences between the same clock
    ctlr_rst_in        => ctlr_rst,
    
    -- Driver clock domain
    dvr_clk            => dvr_clk,
    dvr_rst            => dvr_rst,    
    dvr_en             => dvr_en,
    dvr_wr_not_rd      => dvr_wr_not_rd,
    dvr_done           => dvr_done,
    dvr_start_addr     => dvr_start_addr,
    dvr_end_addr       => dvr_end_addr,
    dvr_wr_flush_en    => dvr_wr_flush_en,
    
    -- Write FIFO clock domain
    wr_clk             => dp_clk,
    wr_rst             => dp_rst,

    wr_fifo_usedw      => wr_fifo_usedw,
    wr_sosi            => wr_src_out, 
    wr_siso            => diag_wr_src_in,
  
    -- Read FIFO clock domain
    rd_clk             => dp_clk,
    rd_rst             => dp_rst,

    rd_fifo_usedw      => rd_fifo_usedw,
    rd_sosi            => diag_rd_snk_in,
    rd_siso            => diag_rd_snk_out,

    -- DDR PHY external interface
    phy_ou             => phy_ou,
    phy_io             => phy_io,
    phy_in             => phy_in
  );
 
  u_tech_ddr_memory_model : ENTITY tech_ddr_lib.tech_ddr_memory_model
  GENERIC MAP (
    g_tech_ddr => c_tech_ddr
  )
  PORT MAP (
    mem_in => phy_ou,
    mem_io => phy_io
  );
    
END ARCHITECTURE str;


