-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.eth_pkg.ALL;

ENTITY eth_control IS
  PORT (
    -- Clocks and reset
    rst               : IN  STD_LOGIC;
    clk               : IN  STD_LOGIC;
    
    -- Control registers
    reg_config        : IN  t_eth_mm_reg_config;
    reg_control       : IN  t_eth_mm_reg_control;
    reg_continue_wr   : IN  STD_LOGIC;     -- used to know that control reg has been written and ETH module can continue from TX_PENDING state
    reg_status        : OUT t_eth_mm_reg_status;
    reg_status_wr     : IN  STD_LOGIC;     -- used to know that status reg has been read and can be cleared to remove the sla_interrupt
    
    -- Streaming sink Rx frame
    -- . The rcv_rd, rcv_ack and rcv_done act instead of rcv_out.ready to have rcv_out ready per frame instead of per data word
    rcv_rd            : OUT STD_LOGIC;
    rcv_ack           : IN  STD_LOGIC;
    rcv_done          : IN  STD_LOGIC;
    rcv_in            : IN  t_dp_sosi;  -- packet memory is always ready for packet data
    rcv_hdr_words_arr : IN  t_network_total_header_32b_arr;
    rcv_hdr_status    : IN  t_eth_hdr_status;
    
    -- Streaming source Tx frame
    xmt_in            : IN  t_dp_siso;
    xmt_out           : OUT t_dp_sosi;
    
    -- MM frame memory    
    mem_in            : OUT t_mem_mosi;
    mem_out           : IN  t_mem_miso
  );
END eth_control;


ARCHITECTURE rtl OF eth_control IS

  -- Internal source ready latency of this component
  CONSTANT c_this_src_latency : NATURAL := 1;  -- xmt_in, xmt_out
  
  TYPE t_state_enum IS (
    s_idle,
    s_rx_request,
    s_rx_frame,
    s_hdr_response,
    s_tx_pending,
    s_tx_frame,
    s_tx_done
  );
  
  SIGNAL nxt_rcv_rd               : STD_LOGIC;
  SIGNAL rcv_in_eop_dly           : STD_LOGIC;
  
  SIGNAL hdr_response_arr         : t_network_total_header_32b_arr;
  SIGNAL nxt_hdr_response_arr     : t_network_total_header_32b_arr;
  SIGNAL hdr_word_cnt             : NATURAL RANGE 0 TO c_network_total_header_32b_nof_words := c_network_total_header_32b_nof_words;
  SIGNAL nxt_hdr_word_cnt         : NATURAL;
  SIGNAL hdr_en                   : STD_LOGIC;
  SIGNAL nxt_hdr_en               : STD_LOGIC;
  SIGNAL hdr_done                 : STD_LOGIC;
  SIGNAL nxt_hdr_done             : STD_LOGIC;
  
  SIGNAL tx_insert                : STD_LOGIC;
  SIGNAL nxt_tx_insert            : STD_LOGIC;
  
  SIGNAL i_reg_status             : t_eth_mm_reg_status;
  SIGNAL nxt_reg_status           : t_eth_mm_reg_status;
  
  SIGNAL xmt_start                : STD_LOGIC;
  SIGNAL nxt_xmt_start            : STD_LOGIC;
  SIGNAL xmt_en                   : STD_LOGIC;
  SIGNAL nxt_xmt_en               : STD_LOGIC;
  SIGNAL xmt_done                 : STD_LOGIC;
  SIGNAL nxt_xmt_done             : STD_LOGIC;
  SIGNAL xmt_word_cnt             : NATURAL RANGE 0 TO c_eth_frame_nof_words;
  SIGNAL nxt_xmt_word_cnt         : NATURAL;
  SIGNAL tx_nof_words             : STD_LOGIC_VECTOR(c_eth_frame_nof_words_w-1 DOWNTO 0);
  SIGNAL nxt_tx_nof_words         : STD_LOGIC_VECTOR(c_eth_frame_nof_words_w-1 DOWNTO 0);
  SIGNAL tx_empty                 : STD_LOGIC_VECTOR(c_eth_empty_w-1 DOWNTO 0);
  SIGNAL nxt_tx_empty             : STD_LOGIC_VECTOR(c_eth_empty_w-1 DOWNTO 0);
  
  SIGNAL i_mem_in                 : t_mem_mosi;
  SIGNAL nxt_mem_in               : t_mem_mosi;
  SIGNAL rd_val                   : STD_LOGIC_VECTOR(0 TO c_mem_ram_rd_latency);  -- use [0] to combinatorially store rd (= rd_en)
  SIGNAL nxt_rd_val               : STD_LOGIC_VECTOR(1 TO c_mem_ram_rd_latency);
  SIGNAL rd_sop                   : STD_LOGIC_VECTOR(    rd_val'RANGE);
  SIGNAL nxt_rd_sop               : STD_LOGIC_VECTOR(nxt_rd_val'RANGE);
  SIGNAL rd_eop                   : STD_LOGIC_VECTOR(    rd_val'RANGE);
  SIGNAL nxt_rd_eop               : STD_LOGIC_VECTOR(nxt_rd_val'RANGE);
  
  SIGNAL xmt_siso                 : t_dp_siso;
  SIGNAL xmt_sosi                 : t_dp_sosi;
  
  SIGNAL i_xmt_out                : t_dp_sosi;
    
  -- ST->MM state machine
  SIGNAL state                    : t_state_enum;
  SIGNAL nxt_state                : t_state_enum;
  
BEGIN

  reg_status <= i_reg_status;
  mem_in     <= i_mem_in;
  xmt_out    <= i_xmt_out;
  
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      rcv_rd                   <= '0';
      rcv_in_eop_dly           <= '0';
      hdr_word_cnt             <= c_network_total_header_32b_nof_words;
      hdr_en                   <= '0';
      hdr_done                 <= '0';
      tx_insert                <= '0';
      xmt_start                <= '0';
      i_reg_status             <= c_eth_mm_reg_status_rst;
      tx_nof_words             <= (OTHERS=>'0');
      tx_empty                 <= (OTHERS=>'0');
      xmt_word_cnt             <= 0;
      xmt_en                   <= '0';
      xmt_done                 <= '0';
      i_mem_in                 <= c_mem_mosi_rst;
      rd_val(nxt_rd_val'RANGE) <= (OTHERS=>'0');
      rd_sop(nxt_rd_val'RANGE) <= (OTHERS=>'0');
      rd_eop(nxt_rd_val'RANGE) <= (OTHERS=>'0');
      state                    <= s_idle;
    ELSIF rising_edge(clk) THEN
      rcv_rd                   <= nxt_rcv_rd;
      rcv_in_eop_dly           <= rcv_in.eop;
      hdr_response_arr         <= nxt_hdr_response_arr;
      hdr_word_cnt             <= nxt_hdr_word_cnt;
      hdr_en                   <= nxt_hdr_en;
      hdr_done                 <= nxt_hdr_done;
      tx_insert                <= nxt_tx_insert;
      xmt_start                <= nxt_xmt_start;
      i_reg_status             <= nxt_reg_status;
      tx_nof_words             <= nxt_tx_nof_words;
      tx_empty                 <= nxt_tx_empty;
      xmt_word_cnt             <= nxt_xmt_word_cnt;
      xmt_en                   <= nxt_xmt_en;
      xmt_done                 <= nxt_xmt_done;
      i_mem_in                 <= nxt_mem_in;
      rd_val(nxt_rd_val'RANGE) <= nxt_rd_val;
      rd_sop(nxt_rd_val'RANGE) <= nxt_rd_sop;
      rd_eop(nxt_rd_val'RANGE) <= nxt_rd_eop;
      state                    <= nxt_state;
    END IF;
  END PROCESS;
  
  ------------------------------------------------------------------------------
  -- Response header
  ------------------------------------------------------------------------------
  
  p_hdr_response_arr : PROCESS(rcv_hdr_words_arr, rcv_hdr_status, reg_config)
  BEGIN
    nxt_hdr_response_arr   <= func_network_total_header_response_eth(     rcv_hdr_words_arr, reg_config.mac_address);
    IF rcv_hdr_status.is_arp='1' THEN
      nxt_hdr_response_arr <= func_network_total_header_response_arp( rcv_hdr_words_arr, reg_config.mac_address, reg_config.ip_address);
    ELSIF rcv_hdr_status.is_icmp='1' THEN
      nxt_hdr_response_arr <= func_network_total_header_response_icmp(rcv_hdr_words_arr, reg_config.mac_address);
    ELSIF rcv_hdr_status.is_udp='1' THEN
      nxt_hdr_response_arr <= func_network_total_header_response_udp( rcv_hdr_words_arr, reg_config.mac_address);
    ELSIF rcv_hdr_status.is_ip='1' THEN
      -- If it is another protocol over IP than ICMP or UDP, then prepare only IP response
      nxt_hdr_response_arr <= func_network_total_header_response_ip(  rcv_hdr_words_arr, reg_config.mac_address);
    END IF;
  END PROCESS;
  
  ------------------------------------------------------------------------------
  -- Frame memory access
  ------------------------------------------------------------------------------
  
  p_mem_access : PROCESS(i_mem_in, rcv_in, hdr_en, hdr_word_cnt, hdr_response_arr, xmt_en, xmt_siso, xmt_word_cnt)
  BEGIN
    nxt_mem_in.address <= i_mem_in.address;
    nxt_mem_in.wrdata  <= i_mem_in.wrdata;
    nxt_mem_in.wr      <= '0';
    nxt_mem_in.rd      <= '0';
    nxt_hdr_word_cnt   <= 0;
    nxt_xmt_word_cnt   <= 0;
    -- Rx frame
    IF rcv_in.valid='1' THEN
      nxt_mem_in.address <= INCR_UVEC(i_mem_in.address, 1);
      nxt_mem_in.wrdata  <= RESIZE_MEM_DATA(rcv_in.data);
      nxt_mem_in.wr      <= '1';
    END IF;
    IF rcv_in.sop='1' THEN
      nxt_mem_in.address <= TO_MEM_ADDRESS(c_eth_ram_rx_offset);  -- rx buffer starts at address 0
    END IF;
    -- Prepare the Tx header
    IF hdr_en='1' THEN                                               -- tx buffer starts at address 380
      nxt_mem_in.address <= TO_MEM_ADDRESS(c_eth_ram_tx_offset + hdr_word_cnt);
      nxt_mem_in.wrdata  <= RESIZE_MEM_DATA(hdr_response_arr(hdr_word_cnt));
      nxt_mem_in.wr      <= '1';
      nxt_hdr_word_cnt   <= hdr_word_cnt+1;
    END IF;
    -- Tx frame
    IF xmt_en='1' THEN
      nxt_mem_in.address <= TO_MEM_ADDRESS(c_eth_ram_tx_offset + xmt_word_cnt);
      nxt_xmt_word_cnt   <= xmt_word_cnt;
      IF xmt_siso.ready='1' THEN
        nxt_mem_in.rd    <= '1';
        nxt_xmt_word_cnt <= xmt_word_cnt+1;  -- rd_en side counter
      END IF;
    END IF;
  END PROCESS;
  
    
  ------------------------------------------------------------------------------
  -- Tx frame output control
  ------------------------------------------------------------------------------
  
  rd_val(0) <= i_mem_in.rd;
  rd_sop(0) <= i_mem_in.rd WHEN xmt_word_cnt=1                      ELSE '0';
  rd_eop(0) <= i_mem_in.rd WHEN xmt_word_cnt=UNSIGNED(tx_nof_words) ELSE '0';
  
  -- rd_val side after read latency
  nxt_rd_val <= rd_val(0 TO rd_val'HIGH-1);
  nxt_rd_sop <= rd_sop(0 TO rd_val'HIGH-1);
  nxt_rd_eop <= rd_eop(0 TO rd_val'HIGH-1);

  xmt_sosi.data  <= RESIZE_DP_DATA(mem_out.rddata);
  xmt_sosi.valid <= rd_val(rd_val'HIGH);
  xmt_sosi.sop   <= rd_sop(rd_val'HIGH);
  xmt_sosi.eop   <= rd_eop(rd_val'HIGH);
  xmt_sosi.empty <= RESIZE_DP_EMPTY(tx_empty);
  
  -- adapt mem read latency to xmt ready latency
  u_xmt_out_adapter : ENTITY dp_lib.dp_latency_adapter
  GENERIC MAP (
    g_in_latency   => c_mem_ram_rd_latency+1,  -- = 2 + 1 latency for xmt_siso.ready -> nxt_mem_in.rd
    g_out_latency  => c_this_src_latency       -- = 1
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    -- ST sink
    snk_out => xmt_siso,
    snk_in  => xmt_sosi,
    -- ST source
    src_in  => xmt_in,
    src_out => i_xmt_out
  );
  
  nxt_xmt_done <= i_xmt_out.eop;

  ------------------------------------------------------------------------------
  -- Master-slave handshake control via status register
  ------------------------------------------------------------------------------
  
  -- Gather Rx and Tx status into the status register
  p_reg_status : PROCESS(i_reg_status, reg_status_wr, rcv_in, rcv_in_eop_dly, i_mem_in, tx_insert, hdr_done, xmt_done)
  BEGIN
    nxt_reg_status <= i_reg_status;
    -- MM write access to the status address clears the status register, which removes the interrupt
    -- using write instead of read access to remove the interrupt allows for polling reg_status
    IF reg_status_wr='1'    THEN nxt_reg_status          <= c_eth_mm_reg_status_rst; END IF;
    -- Events that set the status register
    IF rcv_in.eop='1'       THEN nxt_reg_status.rx_empty <= rcv_in.empty(c_eth_empty_w-1 DOWNTO 0); END IF;
    IF rcv_in_eop_dly='1'   THEN
      nxt_reg_status.rx_nof_words                                     <= (OTHERS=>'0');
      nxt_reg_status.rx_nof_words(c_eth_frame_nof_words_w-1 DOWNTO 0) <= INCR_UVEC(i_mem_in.address(c_eth_frame_nof_words_w-1 DOWNTO 0), 1);
    END IF;
    IF tx_insert='1'        THEN nxt_reg_status.tx_avail <= '1'; END IF;
    IF hdr_done='1'         THEN nxt_reg_status.rx_avail <= '1'; END IF;
    IF xmt_done='1'         THEN nxt_reg_status.tx_done  <= '1'; END IF;
  END PROCESS;
  
  -- Capture the control register tx fields for a new transmit
  nxt_tx_nof_words <= reg_control.tx_nof_words(c_eth_frame_nof_words_w-1 DOWNTO 0) WHEN xmt_start='1' ELSE tx_nof_words;
  nxt_tx_empty     <= reg_control.tx_empty                                         WHEN xmt_start='1' ELSE tx_empty;
  
  
  ------------------------------------------------------------------------------
  -- State machine
  ------------------------------------------------------------------------------
  
  p_state : PROCESS(state, reg_control, reg_continue_wr, rcv_in, rcv_ack, rcv_done, hdr_word_cnt, xmt_siso, xmt_word_cnt, tx_nof_words, xmt_done)
  BEGIN
    nxt_rcv_rd     <= '0';  -- read frame start pulse
    nxt_hdr_en     <= '0';  -- prepare response header enable
    nxt_hdr_done   <= '0';  -- prepare response header done
    nxt_tx_insert  <= '0';  -- accept tx insert request
    nxt_xmt_start  <= '0';  -- write frame start pulse
    nxt_xmt_en     <= '0';  -- write frame enable
    
    nxt_state <= state;
    CASE state IS
      WHEN s_idle =>
        IF reg_control.tx_request='1' THEN   -- MM master requests to insert an extra tx frame
          nxt_tx_insert <= '1';
          nxt_state <= s_tx_pending;
        END IF;
        IF reg_control.rx_en='1' THEN    -- MM master enables default Rx-Tx operation so request a new rx frame
          nxt_rcv_rd <= '1';
          nxt_state <= s_rx_request;
        END IF;
      WHEN s_rx_request =>
        IF rcv_in.sop='1' THEN              -- busy receiving a new frame (best check for sop, so no need to check rcv_busy, rcv_err in eth_rx_frame)
          nxt_state <= s_rx_frame;
        ELSIF rcv_ack='1' THEN              -- rcv ack with no rcv sop, so there was no rx frame pending, try request again
          nxt_state <= s_idle;
        END IF;
      WHEN s_rx_frame =>
        IF rcv_done='1' THEN                -- frame received
          nxt_state <= s_hdr_response;
        END IF;
      WHEN s_hdr_response =>                -- prepare tx header for rx response
        IF hdr_word_cnt < c_network_total_header_32b_nof_words-1 THEN
          nxt_hdr_en <= '1';
        ELSE
          nxt_hdr_done <= '1';              -- new frame received and response prepared
          nxt_state <= s_tx_pending;
        END IF;
      WHEN s_tx_pending =>                  -- wait for MM master
        IF reg_continue_wr='1' THEN      -- MM master done
          IF reg_control.tx_en='1' THEN  -- continue with response tx frame
            nxt_xmt_start <= '1';
            nxt_state <= s_tx_frame;
          ELSE                              -- do not response tx frame
            nxt_state <= s_idle;
          END IF;
        END IF;
      WHEN s_tx_frame =>
        IF xmt_siso.ready='0' OR xmt_word_cnt < UNSIGNED(tx_nof_words)-1 THEN
          nxt_xmt_en <= '1';
        ELSE
          nxt_state <= s_tx_done;           -- frame transmitted
        END IF;
      WHEN OTHERS => -- s_tx_done
        IF xmt_done='1' THEN
          nxt_state <= s_idle;              -- frame transmitted, to be safe wait for xmt_done
        END IF;
    END CASE;
  END PROCESS;
  
END rtl;
