--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, mdio_lib, technology_lib, tech_xaui_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE mdio_lib.mdio_pkg.ALL;
USE mdio_lib.mdio_vitesse_vsc8486_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY tr_xaui IS
  GENERIC (
    g_technology            : NATURAL := c_tech_select_default;
    g_sim                   : BOOLEAN := FALSE;
    g_sim_level             : NATURAL := 0;     -- 0 = use IP; 1 = use fast serdes model
    g_use_xgmii             : BOOLEAN := FALSE; -- Don't use streaming I/O but XGMII (e.g. conenct to 10GbE MAC)
    g_nof_xaui              : NATURAL := 1;     -- Up to 3 (hard XAUI only) supported
    g_use_mdio              : BOOLEAN := FALSE;
    g_mdio_epcs_dis         : BOOLEAN := FALSE -- TRUE disables EPCS on init; e.g. to target a 10GbE card in PC that does not support it 
  );
  PORT (   
    -- Transceiver PLL reference clock   
    tr_clk                  : IN  STD_LOGIC;  -- 156.25 MHz
    tr_rst                  : IN  STD_LOGIC;

    -- Calibration & reconfiguration clock
    cal_rec_clk             : IN  STD_LOGIC;
 
    -- MM interface
    mm_clk                  : IN  STD_LOGIC := '0';
    mm_rst                  : IN  STD_LOGIC := '0';

    xaui_mosi               : IN  t_mem_mosi := c_mem_mosi_rst;
    xaui_miso               : OUT t_mem_miso; 
    
    mdio_mosi_arr           : IN  t_mem_mosi_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>c_mem_mosi_rst);
    mdio_miso_arr           : OUT t_mem_miso_arr(g_nof_xaui-1 DOWNTO 0);

    -- Streaming TX interfaces
    tx_clk_arr              : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);  -- 156.25 MHz, externally connect tr_clk to tx_clk_arr or connect rx_clk_arr to tx_clk_arr
    tx_rst_arr              : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);  -- tx_rst release depends on XAUI ready
    tx_sosi_arr             : IN  t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>c_dp_sosi_rst);
    tx_siso_arr             : OUT t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);

    -- Streaming RX interfaces
    rx_clk_arr_out          : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);  -- recovered clock per XAUI
    rx_clk_arr_in           : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);  -- externally connect to rx_clk_arr_out to avoid clock delta-delay
    rx_rst_arr              : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    rx_sosi_arr             : OUT t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);
    rx_siso_arr             : IN  t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rst);

    -- XGMII interface
    xgmii_tx_dc_arr         : IN  t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>(OTHERS=>'0'));
    xgmii_rx_dc_arr         : OUT t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);
     
    -- Serial IO interface
    xaui_tx_arr             : OUT t_xaui_arr(g_nof_xaui-1 DOWNTO 0);   
    xaui_rx_arr             : IN  t_xaui_arr(g_nof_xaui-1 DOWNTO 0);

    -- MDIO interface
    mdio_rst                : OUT STD_LOGIC;
    mdio_mdc_arr            : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);       
    mdio_mdat_in_arr        : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>'0');
    mdio_mdat_oen_arr       : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0)
  );
END tr_xaui;


ARCHITECTURE str OF tr_xaui IS  

  SIGNAL i_tx_rst_arr              : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL i_rx_rst_arr              : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  --XGMII data and control combined:
  SIGNAL xgmii_tx_dc_in_arr        : t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xgmii_rx_dc_out_arr       : t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);

  --XGMII control bits (one for each XGMII lane):
  SIGNAL xgmii_tx_c_arr            : t_xgmii_c_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xgmii_rx_c_arr            : t_xgmii_c_arr(g_nof_xaui-1 DOWNTO 0);
  
  --XGMII data
  SIGNAL xgmii_tx_d_arr            : t_xgmii_d_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xgmii_rx_d_arr            : t_xgmii_d_arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL txc_tx_ready_arr              : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL rxc_rx_ready_arr              : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL txc_rx_channelaligned_arr     : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL tx_framer_siso_arr        : t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL tx_framer_sosi_arr        : t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);

BEGIN

  tx_rst_arr <= i_tx_rst_arr;
  rx_rst_arr <= i_rx_rst_arr;
  
  -----------------------------------------------------------------------------
  -- XAUI PHY
  -----------------------------------------------------------------------------
  
  i_tx_rst_arr <= NOT txc_tx_ready_arr;
  i_rx_rst_arr <= NOT rxc_rx_ready_arr;
  
  u_tech_xaui : ENTITY tech_xaui_lib.tech_xaui
  GENERIC MAP (
    g_technology => g_technology,
    g_sim        => g_sim,
    g_sim_level  => g_sim_level,
    g_nof_xaui   => g_nof_xaui 
  )
  PORT MAP (
    tr_clk                    => tr_clk,
    tr_rst                    => tr_rst,

    cal_rec_clk               => cal_rec_clk,
  
    mm_rst                    => mm_rst,  
    mm_clk                    => mm_clk,

    xaui_mosi                 => xaui_mosi,
    xaui_miso                 => xaui_miso,
    
    tx_clk_arr                => tx_clk_arr,
    rx_clk_arr_out            => rx_clk_arr_out,
    rx_clk_arr_in             => rx_clk_arr_in,

    txc_tx_ready_arr          => txc_tx_ready_arr,
    rxc_rx_ready_arr          => rxc_rx_ready_arr,
    txc_rx_channelaligned_arr => txc_rx_channelaligned_arr,

    xgmii_tx_dc_arr           => xgmii_tx_dc_in_arr,
    xgmii_rx_dc_arr           => xgmii_rx_dc_out_arr,

    xaui_tx_arr               => xaui_tx_arr,
    xaui_rx_arr               => xaui_rx_arr
  );

  -----------------------------------------------------------------------------
  -- SOSI-XGMII user interface
  -----------------------------------------------------------------------------
  gen_sosi_io: IF g_use_xgmii=FALSE GENERATE
    gen_nof_user : FOR i IN g_nof_xaui-1 DOWNTO 0 GENERATE  
 
      xgmii_tx_dc_in_arr(i) <= func_xgmii_dc(xgmii_tx_d_arr(i), xgmii_tx_c_arr(i));
  
      xgmii_rx_d_arr(i) <= func_xgmii_d(xgmii_rx_dc_out_arr(i));
      xgmii_rx_c_arr(i) <= func_xgmii_c(xgmii_rx_dc_out_arr(i));
  
      tx_siso_arr(i).ready <= '0' WHEN txc_rx_channelaligned_arr(i)='0' ELSE tx_framer_siso_arr(i).ready;
      tx_siso_arr(i).xon   <= txc_rx_channelaligned_arr(i);
  
      tx_framer_sosi_arr(i) <= tx_sosi_arr(i);
  
      u_tx_framer : ENTITY work.tr_xaui_framer
      GENERIC MAP (
        g_dat_len => sel_a_b(g_sim, 100, 100000),
        g_gap_len => 10
       )
      PORT MAP (
        tx_rst     => i_tx_rst_arr(i),
        tx_clk     => tx_clk_arr(i),
  
        snk_out    => tx_framer_siso_arr(i),
        snk_in     => tx_framer_sosi_arr(i),
  
        xgmii_tx_d => xgmii_tx_d_arr(i),
        xgmii_tx_c => xgmii_tx_c_arr(i)    
      );
  
      u_rx_deframer : ENTITY work.tr_xaui_deframer
      PORT MAP (
        rx_rst     => i_rx_rst_arr(i),
        rx_clk     => rx_clk_arr_in(i),
  
        xgmii_rx_d => xgmii_rx_d_arr(i),
        xgmii_rx_c => xgmii_rx_c_arr(i),
  
        src_out    => rx_sosi_arr(i)
      );
    END GENERATE; -- g_nof_xaui
  END GENERATE;

  ---------------------------------------------------------------------------
  -- Direct XGMII user interface
  ---------------------------------------------------------------------------
  gen_xgmii_io: IF g_use_xgmii=TRUE GENERATE
    xgmii_tx_dc_in_arr <= xgmii_tx_dc_arr;
    xgmii_rx_dc_arr    <= xgmii_rx_dc_out_arr;
  END GENERATE;

  ---------------------------------------------------------------------------
  -- MDIO
  ---------------------------------------------------------------------------
  gen_mdio: IF g_use_mdio = TRUE GENERATE
    u_tr_xaui_mdio : ENTITY work.tr_xaui_mdio
    GENERIC MAP (
      g_sim           => g_sim,
      g_nof_xaui      => g_nof_xaui,
      g_mdio_epcs_dis => g_mdio_epcs_dis
    )
    PORT MAP (
      -- Transceiver PLL reference clock   
      tr_clk            => tr_clk,
      tr_rst            => tr_rst,
    
      -- MM clock for register of optional MDIO master
      mm_clk            => mm_clk,
      mm_rst            => mm_rst,
    
      -- MDIO master = mm slave
      mdio_mosi_arr     => mdio_mosi_arr,
      mdio_miso_arr     => mdio_miso_arr,
    
      -- MDIO External clock and serial data.
      mdio_rst          => mdio_rst,
      mdio_mdc_arr      => mdio_mdc_arr,
      mdio_mdat_in_arr  => mdio_mdat_in_arr,
      mdio_mdat_oen_arr => mdio_mdat_oen_arr
    );
  END GENERATE;
      
END str;

