--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose: Test bench for tr_xaui using diagnostics
-- Description:
-- Usage:
--   > as 10
--   > run -a

LIBRARY IEEE, common_lib, dp_lib, diagnostics_lib, technology_lib;                   
USE IEEE.STD_LOGIC_1164.ALL;    
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_pkg.ALL;

ENTITY tb_tr_xaui IS
  GENERIC (
    g_technology         : NATURAL := c_tech_stratixiv;
    g_tb_end             : BOOLEAN := TRUE;   -- when TRUE then tb_end ends this simulation, else a higher multi-testbench will end the simulation
    g_sim_level          : NATURAL := 0       -- 0 = use IP; 1 = use fast serdes model
  );
  PORT (
    tb_end : OUT STD_LOGIC
  );
END ENTITY tb_tr_xaui;

ARCHITECTURE str of tb_tr_xaui IS

  CONSTANT c_sim                 : BOOLEAN := TRUE;  -- tr_xaui_align_dly has time delay ~ 1 sec, so not suitable for simulation
  
  CONSTANT tr_clk_period         : TIME := 6.4 ns;  -- 156.25 MHz
  CONSTANT tx_clk_period         : TIME := 6.4 ns;  -- 156.25 MHz
  CONSTANT mm_clk_period         : TIME := 25 ns;   -- 40 MHz
  CONSTANT cal_rec_clk_period    : TIME := 25 ns;   -- 40 MHz

  CONSTANT phy_delay             : TIME :=  sel_a_b(g_sim_level>0, tr_clk_period*1, 1 ns);  -- the sim_xaui only works without unit PHY delays
  
  CONSTANT c_nof_streams         : NATURAL := 1;
  CONSTANT c_xgmii_data_w        : NATURAL := 64;
  CONSTANT c_nof_lanes           : NATURAL := 4;

  SIGNAL tr_clk                  : STD_LOGIC := '1';
  SIGNAL tr_rst                  : STD_LOGIC;

  SIGNAL xaui_tx_arr             : STD_LOGIC_VECTOR(c_nof_lanes-1 DOWNTO 0);
  SIGNAL xaui_tx_arr_dly         : STD_LOGIC_VECTOR(c_nof_lanes-1 DOWNTO 0);
  SIGNAL xaui_rx_arr             : STD_LOGIC_VECTOR(c_nof_lanes-1 DOWNTO 0);

  SIGNAL rx_sosi_arr             : t_dp_sosi_arr(c_nof_streams-1 DOWNTO 0);
  SIGNAL rx_siso_arr             : t_dp_siso_arr(c_nof_streams-1 DOWNTO 0);

  SIGNAL tx_sosi_arr             : t_dp_sosi_arr(c_nof_streams-1 DOWNTO 0);
  SIGNAL tx_siso_arr             : t_dp_siso_arr(c_nof_streams-1 DOWNTO 0);

  SIGNAL src_diag_en             : STD_LOGIC_VECTOR(c_nof_streams-1 DOWNTO 0);
  SIGNAL src_val_cnt             : t_slv_32_arr(c_nof_streams-1 DOWNTO 0); 

  SIGNAL snk_diag_en             : STD_LOGIC_VECTOR(c_nof_streams-1 DOWNTO 0);
  SIGNAL snk_diag_res            : STD_LOGIC_VECTOR(c_nof_streams-1 DOWNTO 0);
  SIGNAL snk_diag_res_val        : STD_LOGIC_VECTOR(c_nof_streams-1 DOWNTO 0);
  SIGNAL snk_val_cnt             : t_slv_32_arr(c_nof_streams-1 DOWNTO 0);

  SIGNAL tx_clk                  : STD_LOGIC := '1';
  SIGNAL tx_rst                  : STD_LOGIC;

  SIGNAL rx_clk                  : STD_LOGIC;
  SIGNAL rx_rst                  : STD_LOGIC;

  SIGNAL mm_clk                  : STD_LOGIC := '1';
  SIGNAL mm_rst                  : STD_LOGIC := '0';
  
  SIGNAL cal_rec_clk             : STD_LOGIC := '1';

  SIGNAL link_fault              : STD_LOGIC;
  SIGNAL verify_en               : STD_LOGIC;
  
BEGIN

  -- Duration
  p_tb_end : PROCESS  
  BEGIN
    tb_end <= '0';
    link_fault <= '0';
    
    snk_diag_en <= (OTHERS=>'0');
    src_diag_en <= (OTHERS=>'0');
    verify_en <= '0';
    WAIT FOR 2 us;      -- use e.g. wait for 1800 ns to check that p_verify indeed detects errors
    
    -- verify link
    proc_common_wait_until_low(rx_clk, rx_rst);
    proc_common_wait_some_cycles(rx_clk, 20);
    snk_diag_en <= (OTHERS=>'1');
    proc_common_wait_some_cycles(tx_clk, 20);
    src_diag_en <= (OTHERS=>'1');
    proc_common_wait_some_cycles(rx_clk, 50);
    verify_en <= '1';
    WAIT FOR 5 us;
    
    -- model link fault
    snk_diag_en <= (OTHERS=>'0');
    src_diag_en <= (OTHERS=>'0');
    verify_en <= '0';
    link_fault <= '1';
    WAIT FOR 1 us;

    -- verify link recovery after link fault
    link_fault <= '0';
    WAIT FOR 1 us;
    proc_common_wait_until_low(rx_clk, rx_rst);
    proc_common_wait_some_cycles(rx_clk, 20);
    snk_diag_en <= (OTHERS=>'1');
    proc_common_wait_some_cycles(tx_clk, 20);
    src_diag_en <= (OTHERS=>'1');   
    proc_common_wait_some_cycles(rx_clk, 50);
    verify_en <= '1';
    WAIT FOR 5 us;
    
    -- Stop the simulation
    tb_end <= '1';
    IF g_tb_end=FALSE THEN
      REPORT "Tb Simulation finished." SEVERITY NOTE;
    ELSE
      REPORT "Tb Simulation finished." SEVERITY FAILURE;
    END IF;
    WAIT;
  END PROCESS;
  
  tr_clk  <= NOT tr_clk AFTER tr_clk_period/2;    
  tx_clk  <= NOT tx_clk AFTER tx_clk_period/2;    
  mm_clk  <= NOT mm_clk AFTER mm_clk_period/2;    
  cal_rec_clk  <= NOT cal_rec_clk AFTER cal_rec_clk_period/2; 
  
  tr_rst <= '1', '0' AFTER tr_clk_period*4;
  
  -- Verify and duration
  p_verify : PROCESS (rx_clk)
  BEGIN
    IF rising_edge(rx_clk) THEN
      IF verify_en='1' THEN
        ASSERT UNSIGNED(snk_diag_res) = 0 AND SIGNED(snk_diag_res_val) = -1 REPORT "tb_tr_xaui : Wrong diagnostics result value" SEVERITY ERROR;
      END IF;
    END IF;
  END PROCESS;
  
  -- Loopback:
  xaui_tx_arr_dly <= TRANSPORT xaui_tx_arr AFTER phy_delay;
  xaui_rx_arr <= xaui_tx_arr_dly WHEN link_fault='0' ELSE (OTHERS=>'0');

  u_tr_xaui: ENTITY WORK.tr_xaui 
  GENERIC MAP (
    g_technology       => g_technology,
    g_sim              => c_sim,
    g_sim_level        => g_sim_level,
    g_use_mdio         => TRUE
  )
  PORT MAP (     
    tr_clk             => tr_clk,
    tr_rst             => tr_rst,
    
    cal_rec_clk        => cal_rec_clk,
                        
    mm_clk             => mm_clk,
    mm_rst             => mm_rst,
    
    --Parallel data
    tx_clk_arr(0)      => tx_clk,   
    tx_rst_arr(0)      => tx_rst,    
    tx_sosi_arr        => tx_sosi_arr,   
    tx_siso_arr        => tx_siso_arr,
    
    rx_clk_arr_out(0)  => rx_clk,
    rx_clk_arr_in(0)   => rx_clk,
    rx_rst_arr(0)      => rx_rst,   
    rx_sosi_arr        => rx_sosi_arr,   
    rx_siso_arr        => rx_siso_arr,
        
    --Serial data
    xaui_tx_arr(0)     => xaui_tx_arr, 
    xaui_rx_arr(0)     => xaui_rx_arr    
  );

  u_diagnostics: ENTITY diagnostics_lib.diagnostics 
  GENERIC MAP (
    g_dat_w          => c_xgmii_data_w,
    g_nof_streams    => c_nof_streams,
    g_separate_clk   => TRUE
     ) 
  PORT MAP (
    src_rst(0)       => tx_rst,
    src_clk(0)       => tx_clk,

    snk_rst(0)       => rx_rst,
    snk_clk(0)       => rx_clk,

    snk_out_arr      => rx_siso_arr,
    snk_in_arr       => rx_sosi_arr,
    snk_diag_en      => snk_diag_en,
    snk_diag_md      => (OTHERS=>'1'),
    snk_diag_res     => snk_diag_res,
    snk_diag_res_val => snk_diag_res_val,
    snk_val_cnt      => snk_val_cnt,

    src_out_arr      => tx_sosi_arr,
    src_in_arr       => tx_siso_arr,
    src_diag_en      => src_diag_en,
    src_diag_md      => (OTHERS=>'1'),
    src_val_cnt      => src_val_cnt
  );
  
END ARCHITECTURE str;


