-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;

PACKAGE tech_ddr_pkg IS  
  
  -- Gather all parameters in one record
  TYPE t_c_tech_ddr IS RECORD
    -- PHY variant within a technology
    name                              : STRING(1 TO 4);  -- = "DDR3" or "DDR4"
    mts                               : NATURAL;  -- = 800   access rate in mega transfers per second
    master                            : BOOLEAN;  -- = TRUE  TRUE = uniphy master, FALSE = uniphy slave regarding OCT and terminationcontrol for DDR3
    -- PHY external FPGA IO
    a_w                               : NATURAL;  -- = 16
    a_row_w                           : NATURAL;  -- = 16     = a_w, row address width, via a_w lines
    a_col_w                           : NATURAL;  -- = 10    <= a_w, col address width, via a_w lines
    ba_w                              : NATURAL;  -- = 3
    dq_w                              : NATURAL;  -- = 64
    dqs_w                             : NATURAL;  -- = 8      = dq_w / nof_dq_per_dqs
    dm_w                              : NATURAL;  -- = 8
    ck_w                              : NATURAL;  -- = 2
    cke_w                             : NATURAL;  -- = 2
    cs_w                              : NATURAL;  -- = 2      = number of chip select lines
    cs_w_w                            : NATURAL;  -- = 1      = true_log2(cs_w), use when the number of chip select lines is converted to a logical address
    odt_w                             : NATURAL;  -- = 2
    -- PHY internal FPGA IO
    terminationcontrol_w              : NATURAL;  -- = 14     internal bus in FPGA
    -- Controller
    rsl                               : NATURAL;  -- = 4      = 2 (use both PHY clock edges) * 2 (PHY transfer at double rate), resolution
    rsl_w                             : NATURAL;  -- = 2      = ceil_log2(rsl)
    maxburstsize                      : NATURAL;  -- = 64
    maxburstsize_w                    : NATURAL;  -- = 7      = ceil_log2(maxburstsize+1)
  END RECORD;
      
  FUNCTION func_tech_ddr_dq_address_w(  c_ddr : t_c_tech_ddr) RETURN NATURAL;  -- return DDR address width for the DQ data at the PHY mts rate
  FUNCTION func_tech_ddr_ctlr_address_w(c_ddr : t_c_tech_ddr) RETURN NATURAL;  -- return DDR address width for the controller data at the by rsl=4 reduced rate
  FUNCTION func_tech_ddr_ctlr_data_w(   c_ddr : t_c_tech_ddr) RETURN NATURAL;  -- return DDR    data width for the controller data at the by rsl=4 reduced rate
  FUNCTION func_tech_ddr_module_size(   c_ddr : t_c_tech_ddr) RETURN NATURAL;  -- return DDR module size in GByte
  
  CONSTANT c_tech_ddr_max             : t_c_tech_ddr := ("none", 800,  TRUE,   16, 16, 10, 3, 64, 8, 8, 2, 2, 2, 1, 2,   14,   4, 2, 64, 7);  -- maximum ranges for record field definitions
  CONSTANT c_tech_ddr_4g              : t_c_tech_ddr := ("none", 800,  TRUE,   15, 15, 10, 3, 64, 8, 8, 2, 2, 2, 1, 2,   14,   4, 2, 64, 7);
  CONSTANT c_tech_ddr3_4g_800m_master : t_c_tech_ddr := ("DDR3", 800,  TRUE,   15, 15, 10, 3, 64, 8, 8, 2, 2, 2, 1, 2,   14,   4, 2, 64, 7);
  CONSTANT c_tech_ddr3_4g_800m_slave  : t_c_tech_ddr := ("DDR3", 800, FALSE,   15, 15, 10, 3, 64, 8, 8, 2, 2, 2, 1, 2,   14,   4, 2, 64, 7);

  -- PHY in, inout and out signal records
  TYPE t_tech_ddr_phy_in IS RECORD  
    evt                        : STD_LOGIC;                                                          -- event signal is Not Connected to DDR3 PHY
    oct_rup                    : STD_LOGIC;                                                          -- only master DDR3 PHY has On Chip Termination OCT inputs
    oct_rdn                    : STD_LOGIC;                                                          -- only master DDR3 PHY has On Chip Termination OCT inputs
    seriesterminationcontrol   : STD_LOGIC_VECTOR(c_tech_ddr_max.terminationcontrol_w-1 DOWNTO 0);   -- termination control to slave from master DDR3 PHY (internal signal in FPGA)
    parallelterminationcontrol : STD_LOGIC_VECTOR(c_tech_ddr_max.terminationcontrol_w-1 DOWNTO 0);   -- termination control to slave from master DDR3 PHY (internal signal in FPGA)
  END RECORD;
  
  TYPE t_tech_ddr_phy_io IS RECORD
    dq               : STD_LOGIC_VECTOR(c_tech_ddr_max.dq_w-1 DOWNTO 0);   -- data bus
    dqs              : STD_LOGIC_VECTOR(c_tech_ddr_max.dqs_w-1 DOWNTO 0);  -- data strobe bus
    dqs_n            : STD_LOGIC_VECTOR(c_tech_ddr_max.dqs_w-1 DOWNTO 0);
    scl              : STD_LOGIC;                                          -- I2C
    sda              : STD_LOGIC;
  END RECORD;
  
  TYPE t_tech_ddr_phy_ou IS RECORD  
    a                          : STD_LOGIC_VECTOR(c_tech_ddr_max.a_w-1 DOWNTO 0);    -- row and column address
    ba                         : STD_LOGIC_VECTOR(c_tech_ddr_max.ba_w-1 DOWNTO 0);   -- bank address
    dm                         : STD_LOGIC_VECTOR(c_tech_ddr_max.dm_w-1 DOWNTO 0);   -- data mask bus
    cas_n                      : STD_LOGIC;                                          -- column address strobe
    ras_n                      : STD_LOGIC;                                          -- row address strobe
    we_n                       : STD_LOGIC;                                          -- write enable signal
    reset_n                    : STD_LOGIC;                                          -- reset signal
    ck                         : STD_LOGIC_VECTOR(c_tech_ddr_max.ck_w-1 DOWNTO 0);   -- clock, positive edge clock
    ck_n                       : STD_LOGIC_VECTOR(c_tech_ddr_max.ck_w-1 DOWNTO 0);   -- clock, negative edge clock
    cke                        : STD_LOGIC_VECTOR(c_tech_ddr_max.cke_w-1 DOWNTO 0);  -- clock enable
    cs_n                       : STD_LOGIC_VECTOR(c_tech_ddr_max.cs_w-1 DOWNTO 0);   -- chip select
    odt                        : STD_LOGIC_VECTOR(c_tech_ddr_max.odt_w-1 DOWNTO 0);  -- on-die termination control signal
    seriesterminationcontrol   : STD_LOGIC_VECTOR(c_tech_ddr_max.terminationcontrol_w-1 DOWNTO 0);   -- termination control from master to slave DDR3 PHY (internal signal in FPGA)
    parallelterminationcontrol : STD_LOGIC_VECTOR(c_tech_ddr_max.terminationcontrol_w-1 DOWNTO 0);   -- termination control from master to slave DDR3 PHY (internal signal in FPGA)
  END RECORD;

  CONSTANT c_tech_ddr_phy_in_rst   : t_tech_ddr_phy_in := ('0', 'X', 'X', (OTHERS=>'X'), (OTHERS=>'X'));
  CONSTANT c_tech_ddr_phy_io_rst   : t_tech_ddr_phy_io := ((OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), '0', '0');
  CONSTANT c_tech_ddr_phy_ou_rst   : t_tech_ddr_phy_ou := ((OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), '0', '0', '0', '0', (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'));

  TYPE t_tech_ddr_phy_in_arr IS ARRAY(NATURAL RANGE <>) OF t_tech_ddr_phy_in;
  TYPE t_tech_ddr_phy_io_arr IS ARRAY(NATURAL RANGE <>) OF t_tech_ddr_phy_io; 
  TYPE t_tech_ddr_phy_ou_arr IS ARRAY(NATURAL RANGE <>) OF t_tech_ddr_phy_ou;

  -- PHY address signal record
  TYPE t_tech_ddr_addr IS RECORD 
    chip      : STD_LOGIC_VECTOR(c_tech_ddr_max.cs_w_w -1 DOWNTO 0);
    bank      : STD_LOGIC_VECTOR(c_tech_ddr_max.ba_w   -1 DOWNTO 0);
    row       : STD_LOGIC_VECTOR(c_tech_ddr_max.a_row_w-1 DOWNTO 0); 
    col       : STD_LOGIC_VECTOR(c_tech_ddr_max.a_col_w-1 DOWNTO 0);
  END RECORD;
  
  TYPE t_tech_ddr_addr_arr IS ARRAY(NATURAL RANGE <>) OF t_tech_ddr_addr; 
  
  CONSTANT c_tech_ddr_addr_lo     : t_tech_ddr_addr := ((OTHERS=>'0'),
                                                        (OTHERS=>'0'),
                                                        (OTHERS=>'0'),
                                                        (OTHERS=>'0'));

  FUNCTION func_tech_ddr_dq_address(  dq_address : STD_LOGIC_VECTOR; c_ddr : t_c_tech_ddr) RETURN t_tech_ddr_addr;
  FUNCTION func_tech_ddr_dq_address(  ddr_addr   : t_tech_ddr_addr;  c_ddr : t_c_tech_ddr) RETURN STD_LOGIC_VECTOR;
  FUNCTION func_tech_ddr_ctlr_address(ddr_addr   : t_tech_ddr_addr;  c_ddr : t_c_tech_ddr) RETURN STD_LOGIC_VECTOR;
  
  -- PHY MM access signal record
  
  -- Choose smallest maximum slv lengths that fit all use cases, because unconstrained record fields slv is not allowed
  CONSTANT c_tech_ddr_ctlr_address_w    : NATURAL := 32;   -- >= func_tech_ddr_ctlr_address_w(c_tech_ddr_max);
  CONSTANT c_tech_ddr_ctlr_data_w       : NATURAL := 576;  -- >= func_tech_ddr_ctlr_data_w(   c_tech_ddr_max);
  CONSTANT c_tech_ddr_ctlr_burstsize_w  : NATURAL := c_tech_ddr_max.maxburstsize_w;
  
  TYPE t_tech_ddr_miso IS RECORD  
    rddata           : STD_LOGIC_VECTOR(c_tech_ddr_ctlr_data_w-1 DOWNTO 0);
    rdval            : STD_LOGIC;
    waitrequest_n    : STD_LOGIC;
  END RECORD;
  
  TYPE t_tech_ddr_mosi IS RECORD  
    address          : STD_LOGIC_VECTOR(c_tech_ddr_ctlr_address_w-1 DOWNTO 0);
    wrdata           : STD_LOGIC_VECTOR(c_tech_ddr_ctlr_data_w-1 DOWNTO 0);
    wr               : STD_LOGIC;
    rd               : STD_LOGIC;
    burstbegin       : STD_LOGIC;
    burstsize        : STD_LOGIC_VECTOR(c_tech_ddr_ctlr_burstsize_w-1 DOWNTO 0);
  END RECORD;
  
  CONSTANT c_tech_ddr_miso_rst : t_tech_ddr_miso := ((OTHERS=>'0'), '0', '0');
  CONSTANT c_tech_ddr_mosi_rst : t_tech_ddr_mosi := ((OTHERS=>'0'), (OTHERS=>'0'), '0', '0', '0', (OTHERS=>'0'));
  
  -- Resize functions to fit an integer or an SLV in the corresponding t_tech_ddr_miso or t_tech_ddr_mosi field width
  FUNCTION TO_DDR_CTLR_ADDRESS(  n : INTEGER) RETURN STD_LOGIC_VECTOR;  -- unsigned, use integer to support 32 bit range
  FUNCTION TO_DDR_CTLR_DATA(     n : INTEGER) RETURN STD_LOGIC_VECTOR;  -- unsigned
  FUNCTION TO_DDR_CTLR_BURSTSIZE(n : INTEGER) RETURN STD_LOGIC_VECTOR;  -- unsigned

  FUNCTION RESIZE_DDR_CTLR_ADDRESS(  vec : STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR;  -- unsigned
  FUNCTION RESIZE_DDR_CTLR_DATA(     vec : STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR;  -- unsigned
  FUNCTION RESIZE_DDR_CTLR_BURSTSIZE(vec : STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR;  -- unsigned
  
END tech_ddr_pkg;

PACKAGE BODY tech_ddr_pkg IS

  FUNCTION func_tech_ddr_dq_address_w(c_ddr : t_c_tech_ddr) RETURN NATURAL IS
  BEGIN
    RETURN c_ddr.cs_w_w + c_ddr.ba_w + c_ddr.a_w + c_ddr.a_col_w;
  END;
  
  FUNCTION func_tech_ddr_ctlr_address_w(c_ddr : t_c_tech_ddr) RETURN NATURAL IS
    CONSTANT c_dq_address_w : NATURAL := func_tech_ddr_dq_address_w(c_ddr);
  BEGIN
    RETURN c_dq_address_w-c_ddr.rsl_w;
  END;
  
  FUNCTION func_tech_ddr_ctlr_data_w(c_ddr : t_c_tech_ddr) RETURN NATURAL IS
    CONSTANT c_dq_address_w : NATURAL := func_tech_ddr_dq_address_w(c_ddr);
  BEGIN
    RETURN c_ddr.dq_w*c_ddr.rsl;
  END;
  
  FUNCTION func_tech_ddr_module_size(c_ddr : t_c_tech_ddr) RETURN NATURAL IS
    CONSTANT c_dq_address_w       : NATURAL := func_tech_ddr_dq_address_w(c_ddr);
    CONSTANT c_dq_nof_bytes       : NATURAL := 8;  -- both dw_q = 64 and 72 are regarded as having 8 bytes (either with 8 or 9 bits per byte)
    CONSTANT c_dq_nof_bytes_w     : NATURAL := ceil_log2(c_dq_nof_bytes);
    CONSTANT c_module_nof_bytes_w : NATURAL := c_dq_address_w + c_dq_nof_bytes_w;
    CONSTANT c_1GB_w              : NATURAL := 30;
  BEGIN
    RETURN 2**(c_module_nof_bytes_w-c_1GB_w);
  END;
  
  FUNCTION func_tech_ddr_dq_address(dq_address : STD_LOGIC_VECTOR; c_ddr : t_c_tech_ddr) RETURN t_tech_ddr_addr IS
    VARIABLE v_ddr_addr : t_tech_ddr_addr := c_tech_ddr_addr_lo;
  BEGIN
    v_ddr_addr.chip(c_ddr.cs_w_w -1 DOWNTO 0) := dq_address(c_ddr.cs_w_w+c_ddr.ba_w+c_ddr.a_w+c_ddr.a_col_w-1 DOWNTO c_ddr.ba_w+c_ddr.a_w+c_ddr.a_col_w); 
    v_ddr_addr.bank(c_ddr.ba_w   -1 DOWNTO 0) := dq_address(             c_ddr.ba_w+c_ddr.a_w+c_ddr.a_col_w-1 DOWNTO            c_ddr.a_w+c_ddr.a_col_w);
    v_ddr_addr.row( c_ddr.a_w    -1 DOWNTO 0) := dq_address(                        c_ddr.a_w+c_ddr.a_col_w-1 DOWNTO                      c_ddr.a_col_w);
    v_ddr_addr.col( c_ddr.a_col_w-1 DOWNTO 0) := dq_address(                                  c_ddr.a_col_w-1 DOWNTO                                  0);
    RETURN v_ddr_addr;
  END;
    
  FUNCTION func_tech_ddr_dq_address(ddr_addr : t_tech_ddr_addr; c_ddr : t_c_tech_ddr) RETURN STD_LOGIC_VECTOR IS
    CONSTANT c_dq_address_w : NATURAL := func_tech_ddr_dq_address_w(c_ddr);
  BEGIN
    RETURN RESIZE_UVEC(ddr_addr.chip(c_ddr.cs_w_w -1 DOWNTO 0) &
                       ddr_addr.bank(c_ddr.ba_w   -1 DOWNTO 0) &
                       ddr_addr.row( c_ddr.a_row_w-1 DOWNTO 0) &
                       ddr_addr.col( c_ddr.a_col_w-1 DOWNTO 0), c_dq_address_w);
  END;
  
  FUNCTION func_tech_ddr_ctlr_address(ddr_addr : t_tech_ddr_addr; c_ddr : t_c_tech_ddr) RETURN STD_LOGIC_VECTOR IS
    CONSTANT c_dq_address_w   : NATURAL                                     := func_tech_ddr_dq_address_w(c_ddr);
    CONSTANT c_dq_address     : STD_LOGIC_VECTOR(c_dq_address_w-1 DOWNTO 0) := func_tech_ddr_dq_address(ddr_addr, c_ddr);
    CONSTANT c_ctlr_address_w : NATURAL                                     := func_tech_ddr_ctlr_address_w(c_ddr);
  BEGIN
    RETURN RESIZE_UVEC(c_dq_address(c_dq_address_w-1 DOWNTO c_ddr.rsl_w), c_ctlr_address_w);
  END;
  
  FUNCTION TO_DDR_CTLR_ADDRESS(n : INTEGER) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    RETURN RESIZE_UVEC(TO_SVEC(n, 32), c_tech_ddr_ctlr_address_w);
  END TO_DDR_CTLR_ADDRESS;

  FUNCTION TO_DDR_CTLR_DATA(n : INTEGER) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    RETURN RESIZE_UVEC(TO_SVEC(n, 32), c_tech_ddr_ctlr_data_w);
  END TO_DDR_CTLR_DATA;
  
  FUNCTION TO_DDR_CTLR_BURSTSIZE(n : INTEGER) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    RETURN RESIZE_UVEC(TO_SVEC(n, 32), c_tech_ddr_ctlr_burstsize_w);
  END TO_DDR_CTLR_BURSTSIZE;
  
  FUNCTION RESIZE_DDR_CTLR_ADDRESS(vec : STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    RETURN RESIZE_UVEC(vec, c_tech_ddr_ctlr_address_w);
  END RESIZE_DDR_CTLR_ADDRESS;
  
  FUNCTION RESIZE_DDR_CTLR_DATA(vec : STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    RETURN RESIZE_UVEC(vec, c_tech_ddr_ctlr_data_w);
  END RESIZE_DDR_CTLR_DATA;
  
  FUNCTION RESIZE_DDR_CTLR_BURSTSIZE(vec : STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    RETURN RESIZE_UVEC(vec, c_tech_ddr_ctlr_burstsize_w);
  END RESIZE_DDR_CTLR_BURSTSIZE;
  
END tech_ddr_pkg;

