-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Testbench for tech_eth_10g the 10G Ethernet IP technology wrapper.
-- Description:
--   The tb is self checking based on:
--   . proc_tech_mac_10g_rx_packet() for expected header and data type
--   . tx_pkt_cnt=rx_pkt_cnt > 0 must be true at the tb_end.
-- Usage:
--   > as 16
--   > run -all

LIBRARY IEEE, technology_lib, tech_pll_lib, tech_mac_10g_lib, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE tech_mac_10g_lib.tech_mac_10g_component_pkg.ALL;
USE tech_mac_10g_lib.tb_tech_mac_10g_pkg.ALL;
USE tech_pll_lib.tech_pll_component_pkg.ALL;

ENTITY tb_tech_eth_10g IS
  -- Test bench control parameters
  GENERIC (
    g_technology              : NATURAL := c_tech_select_default;
    g_tb_end                  : BOOLEAN := TRUE;   -- when TRUE then tb_end ends this simulation, else a higher multi-testbench will end the simulation
    g_no_dut                  : BOOLEAN := FALSE;  -- default FALSE to verify the DUT, else use TRUE to verify the tb itself without the DUT
    g_sim_level               : NATURAL := 1;      -- 0 = use IP; 1 = use fast serdes model
    g_ref_clk_644_period      : TIME := tech_pll_clk_644_period;  -- for 10GBASE-R
    g_ref_clk_156_period      : TIME := 6.4 ns;                   -- for XAUI
    g_data_type               : NATURAL := c_tb_tech_mac_10g_data_type_symbols;
    g_verify_link_recovery    : BOOLEAN := TRUE;
    g_link_status_check       : STD_LOGIC_VECTOR(c_tech_mac_10g_link_status_w-1 DOWNTO 0) := "11";
    g_use_serial_rx_in        : BOOLEAN := FALSE   -- default FALSE when this tb is ran standalone, else use TRUE to simulate a link between two instances of this tb
  );
  PORT (
    tb_end          : OUT STD_LOGIC;
    -- PHY 10gbase_r
    serial_tx_out   : OUT STD_LOGIC;                                       -- 1 lane
    serial_rx_in    : IN  STD_LOGIC := 'X';
    -- PHY XAUI
    xaui_tx_out     : OUT STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0);   -- 4 lanes
    xaui_rx_in      : IN  STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0) := (OTHERS=>'X')
  );
END tb_tech_eth_10g;


ARCHITECTURE tb OF tb_tech_eth_10g IS

  CONSTANT c_sim                : BOOLEAN := TRUE;  -- tr_xaui_align_dly has time delay ~ 1 sec, so not suitable for simulation
  CONSTANT c_nof_channels       : NATURAL := 1; -- fixed in this tb
  
  CONSTANT cal_clk_period       : TIME := 25 ns;    -- 40 MHz
  
  CONSTANT phy_delay            : TIME :=  0 ns;
  CONSTANT c_phy_loopback       : BOOLEAN := NOT g_use_serial_rx_in;
  
  CONSTANT c_pkt_length_arr1    : t_nat_natural_arr := array_init(0, 50, 1) & (1472, 1473) & 9000;  -- frame longer than 1518-46 = 1472 is received with rx_sosi.err = 8
                                                                                                    -- jumbo frame is 9018-46 = 8972
  CONSTANT c_pkt_length_arr2    : t_nat_natural_arr := array_init(46, 10, 139) & 1472;
  CONSTANT c_pkt_length_arr     : t_nat_natural_arr := c_pkt_length_arr1 & c_pkt_length_arr2;
  CONSTANT c_nof_pkt1           : NATURAL := c_pkt_length_arr1'LENGTH;
  CONSTANT c_nof_pkt2           : NATURAL := c_pkt_length_arr2'LENGTH;
  CONSTANT c_nof_pkt            : NATURAL := sel_a_b(g_verify_link_recovery, c_nof_pkt1 + c_nof_pkt2, c_nof_pkt1);

  CONSTANT c_dst_mac            : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := X"10FA01020300";
  CONSTANT c_src_mac            : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := X"123456789ABC";  -- = 12-34-56-78-9A-BC
  CONSTANT c_src_mac_tx         : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := c_src_mac;
  --CONSTANT c_src_mac_tx         : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := X"100056789ABC";  -- = 10-00-56-78-9A-BC
  CONSTANT c_ethertype          : STD_LOGIC_VECTOR(c_network_eth_type_slv'RANGE) := X"10FA";
  CONSTANT c_etherlen           : STD_LOGIC_VECTOR(c_network_eth_type_slv'RANGE) := "0000000000010000";

  -- Packet headers
  CONSTANT c_eth_header_ethertype : t_network_eth_header := (c_dst_mac, c_src_mac_tx, c_ethertype);
  CONSTANT c_eth_header_etherlen  : t_network_eth_header := (c_dst_mac, c_src_mac_tx, c_etherlen);
  
  SIGNAL total_header      : t_network_total_header := c_network_total_header_ones;  -- default fill all fields with value 1
  
  -- Clocks and reset
  SIGNAL tx_end            : STD_LOGIC := '0';
  SIGNAL rx_end            : STD_LOGIC := '0';
  SIGNAL cal_clk           : STD_LOGIC := '1';  -- calibration clock
  SIGNAL mm_clk            : STD_LOGIC := '0';  -- memory-mapped bus clock
  SIGNAL mm_rst            : STD_LOGIC;         -- reset synchronous with mm_clk
  
  -- External reference clocks
  SIGNAL tr_ref_clk_644    : STD_LOGIC := '1';  -- 10GBASE-R
  SIGNAL tr_ref_clk_312    : STD_LOGIC;         -- 10GBASE-R
  SIGNAL tr_ref_clk_156    : STD_LOGIC := '1';  -- 10GBASE-R or XAUI
  SIGNAL tr_ref_rst_156    : STD_LOGIC;         -- 10GBASE-R or XAUI

  -- XAUI clocks
  SIGNAL tx_rst_arr_out    : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL rx_clk_arr_out    : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL rx_rst_arr_out    : STD_LOGIC_VECTOR(0 DOWNTO 0);
 
  -- . Test bench top level clocks 
  SIGNAL tb_ref_clk_644    : STD_LOGIC;
  SIGNAL tb_ref_clk_312    : STD_LOGIC;
  SIGNAL tb_ref_clk_156    : STD_LOGIC;
  SIGNAL tb_ref_rst_156    : STD_LOGIC;
  SIGNAL tb_tx_clk         : STD_LOGIC;
  SIGNAL tb_tx_rst         : STD_LOGIC;
  SIGNAL tb_rx_clk         : STD_LOGIC;
  SIGNAL tb_rx_rst         : STD_LOGIC;
      
  -- 10G MAC control interface
  SIGNAL mm_init           : STD_LOGIC := '1';
  SIGNAL mac_mosi          : t_mem_mosi;
  SIGNAL mac_mosi_wrdata   : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- 32 bit;
  SIGNAL mac_miso          : t_mem_miso;
  SIGNAL mac_miso_rdval    : STD_LOGIC;
  SIGNAL mac_miso_rddata   : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- 32 bit;
  
  -- 10G MAC transmit interface
  -- . The tb is the ST source
  SIGNAL tx_siso           : t_dp_siso;
  SIGNAL tx_sosi           : t_dp_sosi;
  SIGNAL tx_sosi_data      : STD_LOGIC_VECTOR(c_tech_mac_10g_data_w-1 DOWNTO 0);  -- 64 bit
  
  -- 10G MAC receive interface
  -- . The tb is the ST sink
  SIGNAL rx_siso           : t_dp_siso;
  SIGNAL rx_sosi           : t_dp_sosi;
  SIGNAL rx_sosi_data      : STD_LOGIC_VECTOR(c_tech_mac_10g_data_w-1 DOWNTO 0);  -- 64 bit

  -- PHY serial IO
  -- . 10GBASE-R (single lane)
  SIGNAL serial_tx_arr     : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL serial_rx_arr     : STD_LOGIC_VECTOR(0 DOWNTO 0);
  
  -- . XAUI (four lanes)
  SIGNAL xaui_tx_arr       : t_xaui_arr(0 DOWNTO 0);
  SIGNAL xaui_rx_arr       : t_xaui_arr(0 DOWNTO 0);
      
  -- Model a serial link fault
  SIGNAL link_fault        : STD_LOGIC;

  -- Verification
  SIGNAL tx_pkt_cnt        : NATURAL := 0;
  SIGNAL rx_pkt_cnt        : NATURAL := 0;
  SIGNAL rx_toggle         : STD_LOGIC;  -- toggle after every received packet
  
BEGIN

  cal_clk <= NOT cal_clk AFTER cal_clk_period/2;  -- Calibration clock

  -- debug signals to ease monitoring in wave window  
  mac_mosi_wrdata <= mac_mosi.wrdata(c_word_w-1 DOWNTO 0);
  mac_miso_rddata <= mac_miso.rddata(c_word_w-1 DOWNTO 0);
  mac_miso_rdval <= '1' WHEN mac_mosi.rd='1' AND mac_miso.waitrequest='0' ELSE '0';  -- c_rd_latency = 1
  
  tx_sosi_data <= tx_sosi.data(c_tech_mac_10g_data_w-1 DOWNTO 0);
  rx_sosi_data <= rx_sosi.data(c_tech_mac_10g_data_w-1 DOWNTO 0);
  
  -- Use signal to leave unused fields 'X'
  total_header.eth <= c_eth_header_ethertype;

  u_mm_setup : ENTITY tech_mac_10g_lib.tb_tech_mac_10g_setup
  GENERIC MAP (
    g_technology    => g_technology,
    g_src_mac       => c_src_mac
  )
  PORT MAP (
    tb_end    => rx_end,
    mm_clk    => mm_clk,
    mm_rst    => mm_rst,
    mm_init   => mm_init,
    mac_mosi  => mac_mosi,
    mac_miso  => mac_miso
  );
  
  -- Packet transmitter
  u_transmitter : ENTITY tech_mac_10g_lib.tb_tech_mac_10g_transmitter
  GENERIC MAP (
    g_data_type            => g_data_type,
    g_pkt_length_arr1      => c_pkt_length_arr1,
    g_pkt_length_arr2      => c_pkt_length_arr2,
    g_verify_link_recovery => g_verify_link_recovery
  )
  PORT MAP (
    mm_init        => mm_init,
    total_header   => total_header,
    tx_clk         => tb_tx_clk,
    tx_siso        => tx_siso,
    tx_sosi        => tx_sosi,
    link_fault     => link_fault,
    tx_end         => tx_end
  );
    
  -- Generate reference clocks
  gen_ref_clocks_xaui : IF g_technology=c_tech_stratixiv GENERATE
    tr_ref_clk_644 <= 'X';
    tr_ref_clk_312 <= 'X';
    tr_ref_clk_156 <= NOT tr_ref_clk_156 AFTER g_ref_clk_156_period/2;
    tr_ref_rst_156 <= '1', '0' AFTER g_ref_clk_156_period*5;
  END GENERATE;
    
  gen_ref_clocks_10gbase_r : IF g_technology=c_tech_arria10 GENERATE
    tr_ref_clk_644 <= NOT tr_ref_clk_644 AFTER g_ref_clk_644_period/2;
  
    pll : ENTITY tech_pll_lib.tech_pll_xgmii_mac_clocks
    GENERIC MAP (
      g_technology => g_technology
    )
    PORT MAP (
      refclk_644 => tr_ref_clk_644,
      clk_156    => tr_ref_clk_156,
      clk_312    => tr_ref_clk_312,
      rst_156    => tr_ref_rst_156,
      rst_312    => OPEN
    );
  END GENERATE;
  
  -- Apply the clocks from top level down such that they have their rising_edge() aligned without any delta-delay
  u_tech_eth_10g_clocks : ENTITY work.tech_eth_10g_clocks
  GENERIC MAP (
    g_technology     => g_technology,
    g_nof_channels   => c_nof_channels
  )
  PORT MAP (
    -- Input clocks
    -- . Reference
    tr_ref_clk_644    => tr_ref_clk_644,
    tr_ref_clk_312    => tr_ref_clk_312,
    tr_ref_clk_156    => tr_ref_clk_156,
    tr_ref_rst_156    => tr_ref_rst_156,
    
    -- . XAUI
    tx_rst_arr        => tx_rst_arr_out,
    rx_clk_arr        => rx_clk_arr_out,
    rx_rst_arr        => rx_rst_arr_out,
    
    -- Output clocks
    -- . Reference
    eth_ref_clk_644   => tb_ref_clk_644,
    eth_ref_clk_312   => tb_ref_clk_312,
    eth_ref_clk_156   => tb_ref_clk_156,
    eth_ref_rst_156   => tb_ref_rst_156,
    
    -- . Data
    eth_tx_clk_arr(0) => tb_tx_clk,
    eth_tx_rst_arr(0) => tb_tx_rst,
    
    eth_rx_clk_arr(0) => tb_rx_clk,
    eth_rx_rst_arr(0) => tb_rx_rst
  );  
  
  no_dut : IF g_no_dut=TRUE GENERATE
    tx_rst_arr_out <= (OTHERS=>tr_ref_rst_156);
    rx_clk_arr_out <= (OTHERS=>tr_ref_clk_156);
    rx_rst_arr_out <= (OTHERS=>tr_ref_rst_156);
    
    rx_sosi <= tx_sosi;
    tx_siso <= rx_siso;
  END GENERATE;

  gen_dut : IF g_no_dut=FALSE GENERATE
    dut : ENTITY work.tech_eth_10g
    GENERIC MAP (
      g_technology          => g_technology,
      g_sim                 => c_sim,
      g_sim_level           => g_sim_level,
      g_nof_channels        => c_nof_channels,
      g_link_status_check   => g_link_status_check,
      g_pre_header_padding  => TRUE
    )
    PORT MAP (
      -- Transceiver PLL reference clock
      tr_ref_clk_644    => tb_ref_clk_644,
      tr_ref_clk_312    => tb_ref_clk_312,
      tr_ref_clk_156    => tb_ref_clk_156,
      tr_ref_rst_156    => tb_ref_rst_156,
      
      -- Calibration & reconfig clock
      cal_rec_clk       => cal_clk,         -- for XAUI
      
      -- MM
      mm_clk            => mm_clk,
      mm_rst            => mm_rst,
            
      mac_mosi          => mac_mosi,        -- CSR = control status register
      mac_miso          => mac_miso,
      
      -- XAUI clocks
      tx_clk_arr_in(0)  => tb_tx_clk,       -- 156.25 MHz, externally connect tr_ref_clk_156 to tx_clk_arr or connect rx_clk_arr to tx_clk_arr
      tx_rst_arr_out    => tx_rst_arr_out,
      rx_clk_arr_out    => rx_clk_arr_out,
      rx_clk_arr_in(0)  => tb_rx_clk,       -- externally connect to rx_clk_arr_out to avoid clock delta-delay
      rx_rst_arr_out    => rx_rst_arr_out,
      
      -- ST
      tx_snk_in_arr(0)  => tx_sosi,     -- 64 bit data @ 156 tb_tx_clk
      tx_snk_out_arr(0) => tx_siso,
      
      rx_src_out_arr(0) => rx_sosi,     -- 64 bit data @ 156 tb_rx_clk
      rx_src_in_arr(0)  => rx_siso,
      
      -- PHY serial IO
      -- . 10GBASE-R (single lane)
      serial_tx_arr     => serial_tx_arr,
      serial_rx_arr     => serial_rx_arr,
      
      -- . XAUI (four lanes)
      xaui_tx_arr       => xaui_tx_arr,
      xaui_rx_arr       => xaui_rx_arr
    );
  END GENERATE;

  u_link_connect : ENTITY tech_mac_10g_lib.tb_tech_mac_10g_link_connect
  GENERIC MAP (
    g_loopback    => c_phy_loopback,    -- default TRUE for loopback tx to rx, else use FALSE to connect tx-tx, rx-rx between two tb devices
    g_link_delay  => phy_delay
  )
  PORT MAP (
    link_fault    => link_fault,        -- when '1' then forces rx_serial_arr(0)='0'
    
    -- 10GBASE-R serial layer connect
    serial_tx     => serial_tx_arr(0),
    serial_rx     => serial_rx_arr(0),  -- connects to delayed tx_serial_arr(0) when g_loopback=TRUE else to rx_serial_in
    serial_tx_out => serial_tx_out,     -- connects to delayed tx_serial_arr(0)
    serial_rx_in  => serial_rx_in,      -- used when g_loopback=FALSE
    
    -- XAUI serial layer connect
    xaui_tx       => xaui_tx_arr(0),
    xaui_rx       => xaui_rx_arr(0),    -- connects to delayed xaui_tx_arr(0) when g_loopback=TRUE else to xaui_rx_in
    xaui_tx_out   => xaui_tx_out,       -- connects to delayed xaui_tx_arr(0)
    xaui_rx_in    => xaui_rx_in         -- used when g_loopback=FALSE
  );
  
  -- Packet receiver
  u_receiver : ENTITY tech_mac_10g_lib.tb_tech_mac_10_receiver
  GENERIC MAP (
    g_data_type  => g_data_type
  )
  PORT MAP (
    mm_init        => mm_init,
    total_header   => total_header,
    rx_clk         => tb_rx_clk,
    rx_sosi        => rx_sosi,
    rx_siso        => rx_siso,
    rx_toggle      => rx_toggle
  );
  
  -- Verification
  u_verify_rx_at_eop : ENTITY tech_mac_10g_lib.tb_tech_mac_10_verify_rx_at_eop
  GENERIC MAP (
    g_no_padding     => g_no_dut,
    g_pkt_length_arr => c_pkt_length_arr
  )
  PORT MAP (
    tx_clk      => tb_tx_clk,
    tx_sosi     => tx_sosi,
    rx_clk      => tb_rx_clk,
    rx_sosi     => rx_sosi
  );  

  u_verify_rx_pkt_cnt : ENTITY tech_mac_10g_lib.tb_tech_mac_10g_verify_rx_pkt_cnt
  GENERIC MAP (
    g_nof_pkt   => c_nof_pkt
  )
  PORT MAP (
    tx_clk      => tb_tx_clk,
    tx_sosi     => tx_sosi,
    rx_clk      => tb_rx_clk,
    rx_sosi     => rx_sosi,
    tx_pkt_cnt  => tx_pkt_cnt,
    rx_pkt_cnt  => rx_pkt_cnt,
    rx_end      => rx_end
  );
  
  -- Stop the simulation
  u_simulation_end : ENTITY tech_mac_10g_lib.tb_tech_mac_10g_simulation_end
  GENERIC MAP (
    g_tb_end            => g_tb_end,
    g_nof_clk_to_rx_end => 1000
  )
  PORT MAP (
    clk       => tb_tx_clk,
    tx_end    => tx_end,
    rx_end    => rx_end,
    tb_end    => tb_end
  );
  
END tb;
