-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Testbench for tech_mac_10g the 10G Ethernet IP technology wrapper.
-- Description:
--   The tb is self checking based on:
--   . proc_tech_mac_10g_rx_packet() for expected header and data type
--   . rx_pkt_cnt = tx_pkt_cnt > 0 must be true at the tb_end.
-- Usage:
--   > as 10
--   > run -all

LIBRARY IEEE, technology_lib, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE WORK.tech_mac_10g_component_pkg.ALL;
USE WORK.tb_tech_mac_10g_pkg.ALL;


ENTITY tb_tech_mac_10g IS
  -- Test bench control parameters
  GENERIC (
    g_technology : NATURAL := c_tech_select_default;
    g_tb_end     : BOOLEAN := TRUE;   -- when TRUE then tb_end ends this simulation, else a higher multi-testbench will end the simulation
    g_no_dut     : BOOLEAN := FALSE;  -- default FALSE to verify the DUT, else use TRUE to verify the tb itself without the DUT
    --   g_data_type = c_tb_tech_mac_10g_data_type_symbols  = 0
    --   g_data_type = c_tb_tech_mac_10g_data_type_counter  = 1
    g_data_type  : NATURAL := c_tb_tech_mac_10g_data_type_counter --c_tb_tech_mac_10g_data_type_symbols
  );
  PORT (
    tb_end : OUT STD_LOGIC
  );
END tb_tech_mac_10g;


ARCHITECTURE tb OF tb_tech_mac_10g IS

  CONSTANT clk_156_period       : TIME :=  6.4 ns;  -- 156.25 MHz
  CONSTANT phy_delay            : TIME :=  0 ns;
  
  CONSTANT c_pkt_length_arr     : t_nat_natural_arr := array_init(0, 50, 1) & (1472, 1473) & 9000;  -- frame longer than 1518-46 = 1472 is received with rx_sosi.err = 8
                                                                                                    -- jumbo frame is 9018-46 = 8972
  CONSTANT c_nof_pkt            : NATURAL := c_pkt_length_arr'LENGTH;

  CONSTANT c_dst_mac            : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := X"10FA01020300";
  CONSTANT c_src_mac            : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := X"123456789ABC";  -- = 12-34-56-78-9A-BC
  CONSTANT c_src_mac_tx         : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := c_src_mac;
  --CONSTANT c_src_mac_tx         : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := X"100056789ABC";  -- = 10-00-56-78-9A-BC
  CONSTANT c_ethertype          : STD_LOGIC_VECTOR(c_network_eth_type_slv'RANGE) := X"10FA";
  CONSTANT c_etherlen           : STD_LOGIC_VECTOR(c_network_eth_type_slv'RANGE) := "0000000000010000";

  -- Packet headers
  CONSTANT c_eth_header_ethertype : t_network_eth_header := (c_dst_mac, c_src_mac_tx, c_ethertype);
  CONSTANT c_eth_header_etherlen  : t_network_eth_header := (c_dst_mac, c_src_mac_tx, c_etherlen);
  
  SIGNAL total_header      : t_network_total_header := c_network_total_header_ones;  -- default fill all fields with value 1
  
  -- Clocks and reset
  SIGNAL tx_end            : STD_LOGIC := '0';
  SIGNAL rx_end            : STD_LOGIC := '0';
  SIGNAL mm_clk            : STD_LOGIC := '0';  -- memory-mapped bus clock
  SIGNAL mm_rst            : STD_LOGIC;         -- reset synchronous with mm_clk
  SIGNAL clk_312           : STD_LOGIC := '1';  -- 312.5  MHz
  SIGNAL clk_156           : STD_LOGIC := '0';  -- 156.25 MHz
  SIGNAL tx_ref_clk_312    : STD_LOGIC := '1';
  SIGNAL tx_ref_clk_156    : STD_LOGIC := '0';
  SIGNAL tx_rst            : STD_LOGIC;         -- reset synchronous with tx_ref_clk_156
  SIGNAL rx_phy_clk_312    : STD_LOGIC := '1';
  SIGNAL rx_phy_clk_156    : STD_LOGIC := '0';
  SIGNAL rx_rst            : STD_LOGIC;         -- reset synchronous with rx_phy_clk_156 = tx_ref_clk_156 in this tb

  -- 10G MAC control interface
  SIGNAL mm_init           : STD_LOGIC := '1';
  SIGNAL mac_mosi          : t_mem_mosi;
  SIGNAL mac_mosi_wrdata   : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- 32 bit;
  SIGNAL mac_miso          : t_mem_miso;
  SIGNAL mac_miso_rdval    : STD_LOGIC;
  SIGNAL mac_miso_rddata   : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- 32 bit;
  
  -- 10G MAC transmit interface
  -- . The tb is the ST source
  SIGNAL tx_siso           : t_dp_siso;
  SIGNAL tx_sosi           : t_dp_sosi;
  SIGNAL tx_sosi_data      : STD_LOGIC_VECTOR(c_tech_mac_10g_data_w-1 DOWNTO 0);  -- 64 bit
  
  -- 10G MAC receive interface
  -- . The tb is the ST sink
  SIGNAL rx_siso           : t_dp_siso;
  SIGNAL rx_sosi           : t_dp_sosi;
  SIGNAL rx_sosi_data      : STD_LOGIC_VECTOR(c_tech_mac_10g_data_w-1 DOWNTO 0);  -- 64 bit

  -- 10G MAC XGMII interface
  SIGNAL xgmii_link_status : STD_LOGIC_VECTOR(c_tech_mac_10g_link_status_w-1 DOWNTO 0);  -- 2 bit
  SIGNAL xgmii_tx_data     : STD_LOGIC_VECTOR(c_xgmii_w-1 DOWNTO 0);  -- 72 bit
  SIGNAL xgmii_rx_data     : STD_LOGIC_VECTOR(c_xgmii_w-1 DOWNTO 0);  -- 72 bit

  -- Verification
  SIGNAL tx_pkt_cnt        : NATURAL := 0;
  SIGNAL rx_pkt_cnt        : NATURAL := 0;
  SIGNAL rx_toggle         : STD_LOGIC;  -- toggle after every received packet
  
BEGIN

  clk_156 <= NOT clk_156 AFTER clk_156_period/2;
  clk_312 <= NOT clk_312 AFTER clk_156_period/4;
  tx_ref_clk_156 <= clk_156;               -- mac_10g tx reference clock
  tx_ref_clk_312 <= clk_312;
  rx_phy_clk_156 <= clk_156;               -- use clk_156 to model PHY
  rx_phy_clk_312 <= clk_312;
  
  tx_rst <= '1', '0' AFTER clk_156_period*10;
  rx_rst <= '1', '0' AFTER clk_156_period*10;

  -- debug signals to ease monitoring in wave window  
  mac_mosi_wrdata <= mac_mosi.wrdata(c_word_w-1 DOWNTO 0);
  mac_miso_rddata <= mac_miso.rddata(c_word_w-1 DOWNTO 0);
  mac_miso_rdval <= '1' WHEN mac_mosi.rd='1' AND mac_miso.waitrequest='0' ELSE '0';  -- c_rd_latency = 1
  
  tx_sosi_data <= tx_sosi.data(c_tech_mac_10g_data_w-1 DOWNTO 0);
  rx_sosi_data <= rx_sosi.data(c_tech_mac_10g_data_w-1 DOWNTO 0);
  
  -- Use signal to leave unused fields 'X'
  total_header.eth <= c_eth_header_ethertype;

  u_mm_setup : ENTITY work.tb_tech_mac_10g_setup
  GENERIC MAP (
    g_technology    => g_technology,
    g_src_mac       => c_src_mac
  )
  PORT MAP (
    tb_end    => rx_end,
    mm_clk    => mm_clk,
    mm_rst    => mm_rst,
    mm_init   => mm_init,
    mac_mosi  => mac_mosi,
    mac_miso  => mac_miso
  );

  -- Packet transmitter
  u_transmitter : ENTITY work.tb_tech_mac_10g_transmitter
  GENERIC MAP (
    g_data_type       => g_data_type,
    g_pkt_length_arr1 => c_pkt_length_arr
  )
  PORT MAP (
    mm_init        => mm_init,
    total_header   => total_header,
    tx_clk         => tx_ref_clk_156,
    tx_siso        => tx_siso,
    tx_sosi        => tx_sosi,
    link_fault     => OPEN,
    tx_end         => tx_end
  );
  
  no_dut : IF g_no_dut=TRUE GENERATE
    -- ST loopback
    rx_sosi <= tx_sosi;
    tx_siso <= rx_siso;
  END GENERATE;

  gen_dut : IF g_no_dut=FALSE GENERATE
    dut : ENTITY work.tech_mac_10g
    GENERIC MAP (
      g_technology          => g_technology,
      g_pre_header_padding  => TRUE
    )
    PORT MAP (
      -- MM
      mm_clk            => mm_clk,
      mm_rst            => mm_rst,
      csr_mosi          => mac_mosi,       -- CSR = control status register
      csr_miso          => mac_miso,
  
      -- ST
      tx_clk_312        => tx_ref_clk_312,
      tx_clk_156        => tx_ref_clk_156, -- 156.25 MHz local reference
      tx_rst            => tx_rst,
      tx_snk_in         => tx_sosi,        -- 64 bit data
      tx_snk_out        => tx_siso, 
      
      rx_clk_312        => rx_phy_clk_312,
      rx_clk_156        => rx_phy_clk_156, -- 156.25 MHz from local reference or from rx phy (dependent on g_technology)
      rx_rst            => rx_rst,
      rx_src_out        => rx_sosi,        -- 64 bit data
      rx_src_in         => rx_siso, 
      
      -- XGMII
      xgmii_link_status => xgmii_link_status,
      xgmii_tx_data     => xgmii_tx_data,  -- 72 bit
      xgmii_rx_data     => xgmii_rx_data   -- 72 bit
    );
  END GENERATE;

  -- Loopback XGMII
  u_link_connect : ENTITY work.tb_tech_mac_10g_link_connect
  GENERIC MAP (
    g_loopback    => TRUE,
    g_link_delay  => phy_delay
  )
  PORT MAP (
    -- XGMII layer connect
    xgmii_tx_data => xgmii_tx_data,
    xgmii_rx_data => xgmii_rx_data
  );

  -- Packet receiver
  u_receiver : ENTITY work.tb_tech_mac_10_receiver
  GENERIC MAP (
    g_data_type       => g_data_type
  )
  PORT MAP (
    mm_init        => mm_init,
    total_header   => total_header,
    rx_clk         => rx_phy_clk_156,
    rx_sosi        => rx_sosi,
    rx_siso        => rx_siso,
    rx_toggle      => rx_toggle
  );
  
  -- Verification
  u_verify_rx_at_eop : ENTITY work.tb_tech_mac_10_verify_rx_at_eop
  GENERIC MAP (
    g_no_padding     => g_no_dut,
    g_pkt_length_arr => c_pkt_length_arr
  )
  PORT MAP (
    tx_clk         => tx_ref_clk_156,
    tx_sosi        => tx_sosi,
    rx_clk         => rx_phy_clk_156,
    rx_sosi        => rx_sosi
  );
    
  u_verify_rx_pkt_cnt : ENTITY work.tb_tech_mac_10g_verify_rx_pkt_cnt
  GENERIC MAP (
    g_nof_pkt     => c_nof_pkt
  )
  PORT MAP (
    tx_clk         => tx_ref_clk_156,
    tx_sosi        => tx_sosi,
    rx_clk         => rx_phy_clk_156,
    rx_sosi        => rx_sosi,
    tx_pkt_cnt     => tx_pkt_cnt,
    rx_pkt_cnt     => rx_pkt_cnt,
    rx_end         => rx_end
  );

  -- Stop the simulation
  u_simulation_end : ENTITY work.tb_tech_mac_10g_simulation_end
  GENERIC MAP (
    g_tb_end            => g_tb_end,
    g_nof_clk_to_rx_end => 1000
  )
  PORT MAP (
    clk       => clk_156,
    tx_end    => tx_end,
    rx_end    => rx_end,
    tb_end    => tb_end
  );
  
END tb;
